#! /usr/bin/env python
# -*- coding:utf-8 -*-
"""
@author  : MG
@Time    : 19-4-30 上午9:36
@File    : __init__.py
@contact : mmmaaaggg@163.com
@desc    : 
"""
import os
import pandas as pd
import numpy as np
from ibats_utils.mess import str_2_date

from ibats_common import module_root_path


def load_data(file_name, encoding=None)-> pd.DataFrame:
    file_path = os.path.join(module_root_path, 'example', 'data', file_name)
    df = pd.read_csv(file_path, encoding=encoding)
    return df


def get_trade_date_series():
    df = load_data('trade_date.csv').astype('datetime64[ns]')
    # ret_list = [str_2_date(_) for _ in load_data('trade_date.csv').T.to_numpy()[0]]
    date_s = df.iloc[:, 0]
    return date_s


def get_delivery_date_series(instrument_type):
    df = load_data(
        'future_info.csv'
    ).set_index(
        'symbol'
    ).filter(
        regex='^' + instrument_type + r'(?=\d+$)', axis=0
    )
    # re_pattern_instrument_header = re.compile(r'[A-Za-z]+(?=\d+$)')
    date_s = df["delist_date"].astype('datetime64[ns]').sort_values()
    # ret_list = [str_2_date(_) for _ in load_data('trade_date.csv').T.to_numpy()[0]]
    return date_s


if __name__ == "__main__":
    pass



#! /usr/bin/env python
# -*- coding:utf-8 -*-
"""
@author  : MG
@Time    : 19-5-9 上午9:53
@File    : summary.py
@contact : mmmaaaggg@163.com
@desc    : 
"""
import datetime
import json
import logging
import os
from collections import defaultdict

import docx
import numpy as np
import pandas as pd
from ibats_utils.mess import open_file_with_system_app, date_2_str, datetime_2_str, split_chunk
from scipy.stats import anderson, normaltest

from ibats_common.analysis import get_report_folder_path
from ibats_common.analysis.plot import drawdown_plot, plot_rr_df, wave_hist, plot_scatter_matrix, plot_corr, \
    clean_cache, hist_n_rr, label_distribution
from ibats_common.analysis.plot_db import get_rr_with_md, show_trade, show_cash_and_margin
from ibats_common.backend.mess import get_stg_run_info
from ibats_common.common import RunMode, CalcMode

logger = logging.getLogger(__name__)
STR_FORMAT_DATETIME_4_FILE_NAME = '%Y-%m-%d %H_%M_%S'
FORMAT_2_PERCENT = lambda x: f"{x * 100: .2f}%"
FORMAT_2_FLOAT2 = r"{0:.2f}"
FORMAT_2_FLOAT4 = r"{0:.4f}"


def summary_md(md_df: pd.DataFrame, percentiles=[0.2, 1 / 3, 0.5, 2 / 3, 0.8],
               risk_free=0.03,
               close_key=None,
               enable_show_plot=True,
               enable_save_plot=False,
               name=None,
               func_kwargs_dic={},
               **kwargs,
               ):
    """
    汇总展示数据分析结果，同时以 dict 形式返回各项指标分析结果
    第一个返回值，df的各项分析结果
    第二个返回值，各个列的各项分析结果
    第三个返回值，相关文件输出路径或路径列表（当 enable_save_plot=True）
    :param md_df:
    :param risk_free:无风险收益率
    :param close_key:对哪些列的数据执行统计
    :param enable_show_plot: 展示plot
    :param enable_save_plot: 保存文件
    :param name:
    :param func_kwargs_dic:
    :param kwargs:
    :return:
    """
    columns = list(md_df.columns)
    logger.info('data columns: %s', columns)
    ret_dic, each_col_dic, file_path_dic = {}, defaultdict(dict), {}

    # logger.info('Description:')
    # df.describe(percentiles=percentiles)

    func_kwargs = func_kwargs_dic.setdefault("rr_quantile", None)
    if func_kwargs is not None:
        col_name_list = func_kwargs.setdefault('columns', columns)
        quantile_df = md_df[col_name_list].to_returns().quantile(percentiles).rename(
            columns={_: _ + ' rr' for _ in col_name_list}).T
        ret_dic['rr_quantile'] = quantile_df

    # 获取统计数据
    stats = md_df.calc_stats()
    stats.set_riskfree_rate(risk_free)
    ret_dic['stats'] = stats
    enable_kwargs = {"enable_save_plot": enable_save_plot, "enable_show_plot": enable_show_plot, "name": name}

    # scatter_matrix
    # diagonal，必须且只能在{'hist', 'kde'}中选择1个，
    # 'hist'表示直方图(Histogram plot),'kde'表示核密度估计(Kernel Density Estimation)
    # 该参数是scatter_matrix函数的关键参数
    file_path = plot_scatter_matrix(md_df, diagonal='kde', **enable_kwargs)
    if enable_save_plot:
        file_path_dic['scatter_matrix'] = file_path

    # stats.plot_correlation()
    file_path = plot_corr(md_df, **enable_kwargs)
    if enable_save_plot:
        file_path_dic['correlation'] = file_path

    # return rate plot 图
    func_kwargs = func_kwargs_dic.setdefault('rr', {})
    file_path = plot_rr_df(md_df, **func_kwargs, **enable_kwargs)
    if enable_save_plot:
        file_path_dic['rr'] = file_path

    # histgram 分布图
    func_kwargs = func_kwargs_dic.setdefault('hist', {})
    n_bins_dic, file_path = wave_hist(md_df, **func_kwargs, **enable_kwargs)
    ret_dic['hist'] = n_bins_dic
    if enable_save_plot:
        file_path_dic['hist'] = file_path

    # 回撤图
    func_kwargs = func_kwargs_dic.setdefault('drawdown', {})
    drawdown_df, file_path = drawdown_plot(md_df, perf_stats=stats, **func_kwargs, **enable_kwargs)
    ret_dic['drawdown'] = drawdown_df
    if enable_save_plot:
        file_path_dic['drawdown'] = file_path

    # 未来N日收益率分布
    func_kwargs = func_kwargs_dic.setdefault('hist_future_n_rr', {})
    tmp_dic, file_path = hist_n_rr(md_df, **func_kwargs, **enable_kwargs)
    ret_dic['hist_future_n_rr'] = tmp_dic
    if enable_save_plot:
        file_path_dic['hist_future_n_rr'] = file_path

    file_path_dic['label_distribution'] = defaultdict(dict)
    ret_dic['label_distribution'] = defaultdict(dict)
    noname_enable_kwargs = enable_kwargs.copy()
    del noname_enable_kwargs['name']
    for (n_day, col_name), quantile_df in tmp_dic['quantile_dic'].items():
        tmp_path_dic = file_path_dic['label_distribution'][(n_day, col_name)]
        distribution_dic = ret_dic['label_distribution'][(n_day, col_name)]
        col_count = quantile_df.shape[1]
        for n in range(col_count):
            max_rr = quantile_df.iloc[0, n]
            min_rr = quantile_df.iloc[1, col_count - n - 1]
            distribution_rate_df, file_path = label_distribution(
                md_df[close_key], min_rr=min_rr, max_rr=max_rr, max_future=n_day,
                name=f"{col_name}[{min_rr * 100:.2f}%~{max_rr * 100:.2f}%]", **noname_enable_kwargs)
            tmp_path_dic[(min_rr, max_rr)] = file_path
            distribution_dic[(min_rr, max_rr)] = distribution_rate_df

    # 单列分析
    stat_col_name_list = [close_key]
    stat_df = (md_df if stat_col_name_list is None else md_df[stat_col_name_list])
    for col_name, data in stat_df.items():
        data_df = md_df[[col_name]].dropna()
        data = data.dropna()
        data = data[~np.isinf(data)]
        logger.info("=" * 50 + ' %s ' + "=" * 50, col_name)
        result = normaltest(data)
        each_col_dic[col_name]['normaltest'] = result
        logger.info('%s %s', col_name, result)
        result = anderson(data)
        logger.info('%s %s', col_name, result)
        each_col_dic[col_name]['anderson'] = result
        stats = data.calc_perf_stats()
        # stats = ffn_stats[col_name]
        # 设置无风险收益率
        stats.set_riskfree_rate(risk_free)
        stats.display()
        each_col_dic[col_name]['stats'] = stats
        # plot
        n_bins_dic, file_path = wave_hist(
            md_df[[col_name]].dropna().to_returns().rename(columns={col_name: f'{col_name} rr'}),
            figure_4_each_col=True, name=col_name, **noname_enable_kwargs)
        each_col_dic[col_name]['hist'] = n_bins_dic
        if enable_save_plot:
            file_path_dic[f'{col_name} hist'] = file_path[0]
        rr = data_df.to_returns()
        each_col_dic[col_name]['rr_quantile'] = rr.quantile(
            [0.25, 0.33, 0.5, 0.66, 0.75]
        ).rename(columns={col_name: f"{col_name} rr"}).T

    return ret_dic, each_col_dic, file_path_dic


def summary_rr(df: pd.DataFrame, risk_free=0.03,
               figure_4_each_col=False,
               col_transfer_dic: (dict, None) = None,
               stat_col_name_list=None,
               col_name_list=None,
               enable_show_plot=True,
               enable_save_plot=False,
               name=None,
               **kwargs
               ):
    """
    汇总展示数据分析结果，同时以 dict 形式返回各项指标分析结果
    第一个返回值，df的各项分析结果
    第二个返回值，各个列的各项分析结果
    第三个返回值，相关文件输出路径或路径列表（当 enable_save_plot=True）
    :param df:
    :param risk_free:无风险收益率
    :param figure_4_each_col:hist图使用，每一列显示单独一张图片
    :param col_transfer_dic:列转换方法
    :param stat_col_name_list:对哪些列的数据执行统计
    :param col_name_list:对哪些列的数据执行统计
    :param enable_show_plot: 展示plot
    :param enable_save_plot: 保存文件
    :param name:
    :param kwargs:
    :return:
    """
    columns = list(df.columns)
    logger.info('data columns: %s', columns)
    ret_dic, each_col_dic, file_path_dic = {}, defaultdict(dict), {}

    # 获取统计数据
    stats = df.calc_stats()
    stats.set_riskfree_rate(risk_free)
    ret_dic['stats'] = stats
    enable_kwargs = {"enable_save_plot": enable_save_plot, "enable_show_plot": enable_show_plot, "name": name}

    # scatter_matrix
    # diagonal，必须且只能在{'hist', 'kde'}中选择1个，
    # 'hist'表示直方图(Histogram plot),'kde'表示核密度估计(Kernel Density Estimation)
    # 该参数是scatter_matrix函数的关键参数
    file_path = plot_scatter_matrix(df, diagonal='kde', **enable_kwargs)
    if enable_save_plot:
        file_path_dic['scatter_matrix'] = file_path

    # stats.plot_correlation()
    file_path = plot_corr(df, **enable_kwargs)
    if enable_save_plot:
        file_path_dic['correlation'] = file_path

    # return rate plot 图
    file_path = plot_rr_df(df, **enable_kwargs)
    if enable_save_plot:
        file_path_dic['rr'] = file_path

    # histgram 分布图
    n_bins_dic, file_path = wave_hist(df, figure_4_each_col=figure_4_each_col, col_transfer_dic=col_transfer_dic,
                                      **enable_kwargs)
    ret_dic['hist'] = n_bins_dic
    if enable_save_plot:
        file_path_dic['hist'] = file_path

    # 回撤图
    drawdown_df, file_path = drawdown_plot(df, perf_stats=stats, col_name_list=col_name_list,
                                           **enable_kwargs)
    ret_dic['drawdown'] = drawdown_df
    if enable_save_plot:
        file_path_dic['drawdown'] = file_path

    # 单列分析
    stat_df = (df if stat_col_name_list is None else df[stat_col_name_list])
    for col_name, data in stat_df.items():
        data = data.dropna()
        data = data[~np.isinf(data)]
        logger.info("=" * 50 + ' %s ' + "=" * 50, col_name)
        result = normaltest(data)
        each_col_dic[col_name]['normaltest'] = result
        logger.info('%s %s', col_name, result)
        result = anderson(data)
        logger.info('%s %s', col_name, result)
        each_col_dic[col_name]['anderson'] = result
        stats = data.calc_perf_stats()
        # stats = ffn_stats[col_name]
        # 设置无风险收益率
        stats.set_riskfree_rate(risk_free)
        stats.display()

    return ret_dic, each_col_dic, file_path_dic


def df_2_table(doc, df, format_by_index=None, format_by_col=None, max_col_count=None):
    if max_col_count is None:
        max_col_count = df.shape[1]

    for table_num, col_name_list in enumerate(split_chunk(list(df.columns), max_col_count)):
        if table_num > 0:
            # 如果是换行写入第二、三、四。。个表格，先打一个空行
            doc.add_paragraph('')

        sub_df = df[col_name_list]
        row_num, col_num = sub_df.shape
        t = doc.add_table(row_num + 1, col_num + 1)

        # Highlight all cells limegreen (RGB 32CD32) if cell contains text "0.5"
        from docx.oxml.ns import nsdecls
        from docx.oxml import parse_xml
        from docx.enum.text import WD_ALIGN_PARAGRAPH

        # write head
        # col_name_list = list(sub_df.columns)
        for j in range(col_num):
            # t.cell(0, j).text = df.columns[j]
            # paragraph = t.cell(0, j).add_paragraph()
            paragraph = t.cell(0, j + 1).paragraphs[0]
            paragraph.add_run(str(col_name_list[j])).bold = True
            paragraph.alignment = WD_ALIGN_PARAGRAPH.CENTER

        # write head bg color
        for j in range(col_num + 1):
            # t.cell(0, j).text = df.columns[j]
            t.cell(0, j)._tc.get_or_add_tcPr().append(
                parse_xml(r'<w:shd {} w:fill="00A2E8"/>'.format(nsdecls('w'))))

        # format table style to be a grid
        t.style = 'TableGrid'

        # populate the table with the dataframe
        for i in range(row_num):
            index = sub_df.index[i]
            paragraph = t.cell(i + 1, 0).paragraphs[0]
            index_str = str(date_2_str(index))
            paragraph.add_run(index_str).bold = True
            paragraph.alignment = WD_ALIGN_PARAGRAPH.LEFT
            if format_by_index is not None and index in format_by_index:
                formater = format_by_index[index]
            else:
                formater = None

            for j in range(col_num):
                if formater is None and format_by_col is not None and col_name_list[j] in format_by_col:
                    formater = format_by_col[col_name_list[j]]

                content = sub_df.values[i, j]
                if formater is None:
                    text = str(content)
                elif isinstance(formater, str):
                    text = str.format(formater, content)
                elif callable(formater):
                    text = formater(content)
                else:
                    raise ValueError('%s: %s 无效', index, formater)

                paragraph = t.cell(i + 1, j + 1).paragraphs[0]
                paragraph.alignment = WD_ALIGN_PARAGRAPH.RIGHT
                paragraph.add_run(text)

        for i in range(1, row_num + 1):
            for j in range(col_num + 1):
                if i % 2 == 0:
                    t.cell(i, j)._tc.get_or_add_tcPr().append(
                        parse_xml(r'<w:shd {} w:fill="A3D9EA"/>'.format(nsdecls('w'))))


def _test_summary_md():
    from ibats_common.example.data import load_data

    df = load_data('RB.csv').set_index('trade_date').drop('instrument_type', axis=1)
    df.index = pd.DatetimeIndex(df.index)
    col_transfer_dic = {
        'return': ['open', 'high', 'low', 'close', 'volume']
    }
    ret_dic, each_col_dic, file_path_dic = summary_md(
        df, enable_show_plot=True, enable_save_plot=False,
        func_kwargs_dic={
            "hist": {
                "figure_4_each_col": False,
                "col_transfer_dic": col_transfer_dic,
            },
            "drawdown": {
                "col_name_list": ['close'],
                "stat_col_name_list": ['close'],
            }
        })


def summary_stg(stg_run_id=None):
    from ibats_common.analysis.plot_db import show_order, show_cash_and_margin, show_rr_with_md
    info = get_stg_run_info(stg_run_id)
    stg_run_id = info.stg_run_id

    data_dict, file_path = show_order(stg_run_id)
    df = show_cash_and_margin(stg_run_id)
    sum_df, symbol_rr_dic, save_file_path_dic = show_rr_with_md(stg_run_id)
    summary_rr(sum_df, figure_4_each_col=True, col_transfer_dic={'return': sum_df.columns})
    # for symbol, rr_df in symbol_rr_dic.items():
    #     col_transfer_dic = {'return': rr_df.columns}
    #     summary_rr(rr_df, figure_4_each_col=True, col_transfer_dic=col_transfer_dic)


def _test_summary_stg():
    stg_run_id = None
    summary_stg(stg_run_id)


def stats_df_2_docx_table(stats_df, document):
    """
    将 stats_df 统计信息以表格形式写入 docx 文件中
    :param stats_df:
    :param document:
    :return:
    """

    format_by_index = {
        "total_return": FORMAT_2_PERCENT,
        "cagr": FORMAT_2_PERCENT,
        "mtd": FORMAT_2_PERCENT,
        "three_month": FORMAT_2_PERCENT,
        "six_month": FORMAT_2_PERCENT,
        "ytd": FORMAT_2_PERCENT,
        "one_year": FORMAT_2_PERCENT,
        "three_year": FORMAT_2_PERCENT,
        "five_year": FORMAT_2_PERCENT,
        "ten_year": FORMAT_2_PERCENT,
        "best_day": FORMAT_2_PERCENT,
        "worst_day": FORMAT_2_PERCENT,
        "best_month": FORMAT_2_PERCENT,
        "worst_month": FORMAT_2_PERCENT,
        "best_year": FORMAT_2_PERCENT,
        "worst_year": FORMAT_2_PERCENT,
        "avg_drawdown": FORMAT_2_PERCENT,
        "avg_up_month": FORMAT_2_PERCENT,
        "win_year_perc": FORMAT_2_PERCENT,
        "twelve_month_win_perc": FORMAT_2_PERCENT,
        "incep": FORMAT_2_PERCENT,
        "rf": FORMAT_2_FLOAT2,
        "max_drawdown": FORMAT_2_PERCENT,
        "calmar": FORMAT_2_FLOAT2,
        "daily_sharpe": FORMAT_2_FLOAT2,
        "daily_sortino": FORMAT_2_FLOAT2,
        "daily_mean": FORMAT_2_FLOAT2,
        "daily_vol": FORMAT_2_FLOAT2,
        "daily_skew": FORMAT_2_FLOAT2,
        "daily_kurt": FORMAT_2_FLOAT2,
        "monthly_sharpe": FORMAT_2_FLOAT2,
        "monthly_sortino": FORMAT_2_FLOAT2,
        "monthly_mean": FORMAT_2_FLOAT2,
        "monthly_vol": FORMAT_2_FLOAT2,
        "monthly_skew": FORMAT_2_FLOAT2,
        "monthly_kurt": FORMAT_2_FLOAT2,
        "yearly_sharpe": FORMAT_2_FLOAT2,
        "yearly_sortino": FORMAT_2_FLOAT2,
        "yearly_mean": FORMAT_2_FLOAT2,
        "yearly_vol": FORMAT_2_FLOAT2,
        "yearly_skew": FORMAT_2_FLOAT2,
        "yearly_kurt": FORMAT_2_FLOAT2,
        "avg_drawdown_days": FORMAT_2_FLOAT2,
        "avg_down_month": FORMAT_2_FLOAT2,
        "start": date_2_str,
        "end": date_2_str,
    }
    df_2_table(document, stats_df, format_by_index=format_by_index)


def summary_stg_2_docx(stg_run_id=None, enable_save_plot=True, enable_show_plot=False, enable_clean_cache=True,
                       doc_file_path=None):
    """
    生成策略分析报告
    :param stg_run_id:
    :param enable_save_plot:
    :param enable_show_plot:
    :param enable_clean_cache:
    :param doc_file_path: 可以是目录 或 文件名路径
    :return:
    """
    info = get_stg_run_info(stg_run_id)
    stg_run_id = info.stg_run_id
    run_mode = RunMode(info.run_mode)
    kwargs = {"enable_show_plot": enable_show_plot, "enable_save_plot": enable_save_plot, "run_mode": run_mode}
    if run_mode == RunMode.Backtest_FixPercent:
        sum_df, symbol_rr_dic = get_rr_with_md(stg_run_id, compound_rr=True)
    else:
        sum_df, symbol_rr_dic = get_rr_with_md(stg_run_id, compound_rr=False)

    ret_dic, each_col_dic, file_path_dic = summary_rr(sum_df, **kwargs)

    _, file_path = show_trade(stg_run_id, **kwargs)
    file_path_dic['trade'] = file_path
    df, file_path = show_cash_and_margin(stg_run_id, **kwargs)
    file_path_dic['cash_and_margin'] = file_path

    # 生成 docx 文档将所需变量
    heading_title = f'策略分析报告[{stg_run_id}] ' \
        f'{date_2_str(min(sum_df.index))} - {date_2_str(max(sum_df.index))} ({sum_df.shape[0]} days)'

    # 生成 docx 文件
    document = docx.Document()
    # 设置默认字体
    document.styles['Normal'].font.name = '微软雅黑'
    document.styles['Normal']._element.rPr.rFonts.set(docx.oxml.ns.qn('w:eastAsia'), '微软雅黑')
    # 创建自定义段落样式(第一个参数为样式名, 第二个参数为样式类型, 1为段落样式, 2为字符样式, 3为表格样式)
    UserStyle1 = document.styles.add_style('UserStyle1', 1)
    # 设置字体尺寸
    UserStyle1.font.size = docx.shared.Pt(40)
    # 设置字体颜色
    UserStyle1.font.color.rgb = docx.shared.RGBColor(0xff, 0xde, 0x00)
    # 居中文本
    UserStyle1.paragraph_format.alignment = docx.enum.text.WD_ALIGN_PARAGRAPH.CENTER
    # 设置中文字体
    UserStyle1.font.name = '微软雅黑'
    UserStyle1._element.rPr.rFonts.set(docx.oxml.ns.qn('w:eastAsia'), '微软雅黑')

    # 文件内容
    document.add_heading(heading_title, 0).alignment = docx.enum.text.WD_ALIGN_PARAGRAPH.CENTER
    document.add_paragraph('')
    document.add_paragraph('')
    heading_count = 1
    document.add_heading(f'{heading_count}、策略回测收益曲线', 1)
    # 增加图片（此处使用相对位置）
    document.add_picture(file_path_dic['rr'])  # , width=docx.shared.Inches(1.25)
    heading_count += 1
    # 添加分页符
    document.add_page_break()

    document.add_heading(f'{heading_count}、策略回撤曲线', 1)
    document.add_picture(file_path_dic['drawdown'])
    heading_count += 1
    document.add_page_break()

    if run_mode != RunMode.Backtest_FixPercent:
        document.add_heading(f'{heading_count}、现金与仓位堆叠图', 1)
        document.add_picture(file_path_dic['cash_and_margin'])
        heading_count += 1
        document.add_page_break()

    document.add_heading(f'{heading_count}、散点图矩阵图（Scatter Matrix）', 1)
    document.add_picture(file_path_dic['scatter_matrix'])
    heading_count += 1
    document.add_page_break()

    document.add_heading(f'{heading_count}、相关性矩阵图（Correlation）', 1)
    document.add_picture(file_path_dic['correlation'])
    heading_count += 1
    document.add_page_break()

    document.add_heading(f'{heading_count}、绩效统计数据（Porformance stat）', 1)
    stats_df = ret_dic['stats'].stats
    stats_df_2_docx_table(stats_df, document)
    heading_count += 1
    document.add_page_break()

    # 交易记录
    document.add_heading(f'{heading_count}、买卖点记录', 1)
    document.add_picture(file_path_dic['trade'])
    heading_count += 1
    document.add_page_break()

    # 保存文件
    if doc_file_path is not None:
        if os.path.isdir(doc_file_path):
            folder_path, file_name = doc_file_path, ''
        else:
            folder_path, file_name = os.path.split(doc_file_path)
    else:
        folder_path, file_name = get_report_folder_path(), ''

    if folder_path != '' and not os.path.exists(folder_path):
        os.makedirs(folder_path)

    if file_name == '':
        try:
            calc_mode_str = CalcMode(json.loads(info.trade_agent_params_list)[0]['calc_mode']).name + " "
        except:
            calc_mode_str = " "

        run_mode_str = run_mode.name + " "
        file_name = f"{stg_run_id} {run_mode_str}{calc_mode_str}" \
            f"{date_2_str(min(sum_df.index))} - {date_2_str(max(sum_df.index))} ({sum_df.shape[0]} days) " \
            f"{datetime_2_str(datetime.datetime.now(), STR_FORMAT_DATETIME_4_FILE_NAME)}.docx"
        file_path = os.path.join(folder_path, file_name)
    else:
        file_path = doc_file_path

    document.save(file_path)
    if enable_clean_cache:
        clean_cache()

    return file_path


def _test_summary_stg_2_docx(auto_open_file=True):
    stg_run_id = 1
    file_path = summary_stg_2_docx(stg_run_id, enable_clean_cache=True)
    if auto_open_file:
        open_file_with_system_app(file_path)


def summary_md_2_docx(md_df: pd.DataFrame, percentiles=[0.2, 0.33, 0.5, 0.66, 0.8],
                      risk_free=0.03,
                      close_key=None,
                      enable_show_plot=True,
                      enable_save_plot=False,
                      name=None,
                      func_kwargs_dic={},
                      enable_clean_cache=True,
                      ):
    """
    汇总展示数据分析结果，同时以 dict 形式返回各项指标分析结果
    第一个返回值，df的各项分析结果
    第二个返回值，各个列的各项分析结果
    :param md_df:
    :param percentiles:分为数信息
    :param risk_free:无风险收益率
    :param close_key:对哪些列的数据执行统计
    :param enable_show_plot:显示plot
    :param enable_save_plot:保存plot
    :param name:
    :param func_kwargs_dic:
    :param enable_clean_cache:
    :return:
    """
    ret_dic, each_col_dic, file_path_dic = summary_md(
        md_df, percentiles=percentiles, risk_free=risk_free, close_key=close_key,
        enable_show_plot=enable_show_plot, enable_save_plot=enable_save_plot, name=name,
        func_kwargs_dic=func_kwargs_dic)

    logger.debug('file_path_dic')
    for num, (k, v) in enumerate(file_path_dic.items(), start=1):
        if isinstance(v, dict):
            for num2, (k2, v2) in enumerate(v.items(), start=1):
                if isinstance(v2, dict):
                    for num3, (k3, v3) in enumerate(v2.items(), start=1):
                        logger.debug("%d.%d.%d) %s %s %s -> %s", num, num2, num3, k, k2, k3, v3)
                else:
                    logger.debug("%d.%d) %s %s -> %s", num, num2, k, k2, v2)
        else:
            logger.debug("%d) %s -> %s", num, k, v)

    # 生成 docx 文档将所需变量
    heading_title = f'数据分析报告 {date_2_str(min(md_df.index))} - {date_2_str(max(md_df.index))} ({md_df.shape[0]} days)'

    # 生成 docx 文件
    document = docx.Document()
    # 设置默认字体
    document.styles['Normal'].font.name = '微软雅黑'
    document.styles['Normal']._element.rPr.rFonts.set(docx.oxml.ns.qn('w:eastAsia'), '微软雅黑')
    # 创建自定义段落样式(第一个参数为样式名, 第二个参数为样式类型, 1为段落样式, 2为字符样式, 3为表格样式)
    UserStyle1 = document.styles.add_style('UserStyle1', 1)
    # 设置字体尺寸
    UserStyle1.font.size = docx.shared.Pt(40)
    # 设置字体颜色
    UserStyle1.font.color.rgb = docx.shared.RGBColor(0xff, 0xde, 0x00)
    # 居中文本
    UserStyle1.paragraph_format.alignment = docx.enum.text.WD_ALIGN_PARAGRAPH.CENTER
    # 设置中文字体
    UserStyle1.font.name = '微软雅黑'
    UserStyle1._element.rPr.rFonts.set(docx.oxml.ns.qn('w:eastAsia'), '微软雅黑')

    # 文件内容
    document.add_heading(heading_title, 0).alignment = docx.enum.text.WD_ALIGN_PARAGRAPH.CENTER
    document.add_paragraph('')
    document.add_paragraph('')
    heading_count = 1
    if 'rr' in file_path_dic:
        document.add_heading(f'{heading_count}、行情曲线', 1)
        # 增加图片（此处使用相对位置）
        document.add_picture(file_path_dic['rr'])  # , width=docx.shared.Inches(1.25)
        heading_count += 1
        # 添加分页符
        document.add_page_break()

    if 'hist' in file_path_dic:
        document.add_heading(f'{heading_count}、Histgram 分布图', 1)
        document.add_picture(file_path_dic['hist'])
        heading_count += 1

    if 'rr_quantile' in ret_dic:
        document.add_heading(f'{heading_count}、分位数信息（Quantile）', 1)
        rr_quantile_df = ret_dic['rr_quantile']
        format_by_col = {_: FORMAT_2_PERCENT for _ in rr_quantile_df.columns}
        df_2_table(document, rr_quantile_df, format_by_col=format_by_col, max_col_count=5)
        heading_count += 1
        document.add_page_break()

    if 'hist_future_n_rr' in file_path_dic:
        document.add_heading(f'{heading_count}、未来N日收益率最高最低值分布图', 1)
        quantile_dic = ret_dic['hist_future_n_rr']['quantile_dic']
        for num, ((n_day, col_name), file_path) in enumerate(file_path_dic['hist_future_n_rr'].items(), start=1):
            document.add_heading(f'{heading_count}.{num}) 未来 {n_day} 日 {col_name} 收益率最高最低值分布图', 2)
            document.add_picture(file_path)
            document.add_heading(f'{heading_count}.{num}.1) 分位数信息', 3)
            data_df = quantile_dic[(n_day, col_name)]
            df_2_table(document, data_df, format_by_index={_: FORMAT_2_PERCENT for _ in data_df.index})
            document.add_heading(f'{heading_count}.{num}.2) 三分类标签分布比例', 3)
            for (min_pct, max_pct), distribution_rate_df in ret_dic['label_distribution'][(n_day, col_name)].items():
                file_path = file_path_dic['label_distribution'][(n_day, col_name)][(min_pct, max_pct)]
                document.add_picture(file_path)
                distribution_rate_df.rename(
                    columns={1: f'1 under {min_pct * 100:.2f}%', 2: f'2 over {max_pct * 100:.2f}%'}, inplace=True)
                df_2_table(document, distribution_rate_df,
                           format_by_index={_: FORMAT_2_PERCENT for _ in distribution_rate_df.index})

            document.add_page_break()

        heading_count += 1

    if 'drawdown' in file_path_dic:
        document.add_heading(f'{heading_count}、行情回撤曲线', 1)
        document.add_picture(file_path_dic['drawdown'])
        heading_count += 1
        document.add_page_break()

    if 'scatter_matrix' in file_path_dic:
        document.add_heading(f'{heading_count}、散点图矩阵图（Scatter Matrix）', 1)
        document.add_picture(file_path_dic['scatter_matrix'])
        heading_count += 1
        document.add_page_break()

    if 'correlation' in file_path_dic:
        document.add_heading(f'{heading_count}、相关性矩阵图（Correlation）', 1)
        document.add_picture(file_path_dic['correlation'])
        heading_count += 1
        document.add_page_break()

    if 'stats' in file_path_dic:
        document.add_heading(f'{heading_count}、绩效统计数据（Porformance stat）', 1)
        stats_df = ret_dic['stats'].stats
        stats_df_2_docx_table(stats_df, document)
        heading_count += 1
        document.add_page_break()

    # 逐列分析
    col_name = close_key
    # 文件内容
    heading_title = f'{col_name} 数据分析结果'
    document.add_heading(heading_title, 0).alignment = docx.enum.text.WD_ALIGN_PARAGRAPH.CENTER
    document.add_paragraph('')
    document.add_paragraph('')
    heading_count = 1

    if 'rr_quantile' in ret_dic:
        document.add_heading(f'{heading_count}、分位数信息（Quantile）', 1)
        data_df = each_col_dic[col_name]['rr_quantile']
        format_by_col = {_: FORMAT_2_PERCENT for _ in data_df.columns}
        df_2_table(document, data_df, format_by_col=format_by_col, max_col_count=5)
        heading_count += 1

    if f'{col_name} hist' in file_path_dic:
        document.add_heading(f'{heading_count}、Histgram 分布图', 1)
        document.add_picture(file_path_dic[f'{col_name} hist'])
        heading_count += 1
        document.add_page_break()

    # 保存文件
    file_name = f"MD {date_2_str(min(md_df.index))} - {date_2_str(max(md_df.index))} ({md_df.shape[0]} days) " \
        f"{datetime_2_str(datetime.datetime.now(), STR_FORMAT_DATETIME_4_FILE_NAME)}.docx"
    file_path = os.path.join(get_report_folder_path(), file_name)
    document.save(file_path)
    if enable_clean_cache:
        clean_cache()

    return file_path


def _test_summary_md_2_docx(auto_open_file=True):
    from ibats_common.example.data import load_data
    instrument_type = 'RB'
    file_name = f"{instrument_type}.csv"

    df = load_data(file_name).set_index('trade_date').drop('instrument_type', axis=1)
    df.index = pd.DatetimeIndex(df.index)
    column_list_oraginal = list(df.columns)

    from ibats_common.backend.factor import get_factor
    from ibats_common.example.data import get_trade_date_series
    from ibats_common.example.data import get_delivery_date_series
    df = get_factor(df, close_key='close',
                    trade_date_series=get_trade_date_series(),
                    delivery_date_series=get_delivery_date_series(instrument_type))

    col_transfer_dic = {
        'return': ['open', 'high', 'low', 'close', 'volume']
    }
    file_path = summary_md_2_docx(
        df, enable_show_plot=False, enable_save_plot=True, close_key='close',
        func_kwargs_dic={
            "hist": {
                "figure_4_each_col": False,
                "columns": column_list_oraginal,
                "col_transfer_dic": col_transfer_dic,
            },
            "drawdown": {
                "col_name_list": ['close'],
            },
            "rr": {
                "col_name_list": ['close'],
            },
            "hist_future_n_rr": {
                'n_days': [3, 5],
                "columns": ['close'],
            },
            "rr_quantile": {'columns': ['close']},
            # "": {},
        })
    if auto_open_file:
        open_file_with_system_app(file_path)


if __name__ == "__main__":
    # _test_summary_md()
    # _test_summary_stg()
    # _test_summary_stg_2_docx()
    _test_summary_md_2_docx()



#! /usr/bin/env python
# -*- coding:utf-8 -*-
"""
@author  : MG
@Time    : 2019/5/20 21:09
@File    : factor.py
@contact : mmmaaaggg@163.com
@desc    : 用来对时间序列数据建立因子
"""
import pandas as pd
import logging
import datetime
import numpy as np
import ffn
from ibats_common.example.data import get_delivery_date_series


logger = logging.getLogger(__name__)


def add_factor_of_trade_date(df: pd.DataFrame, trade_date_series):
    """
    自然日，交易日相关因子
    :param df:
    :param trade_date_series:
    :return:
    """
    # 自然日相关因子
    index_s = pd.Series(df.index)
    df['dayofweek'] = index_s.apply(lambda x: x.dayofweek).to_numpy()
    df['day'] = index_s.apply(lambda x: x.day).to_numpy()
    df['month'] = index_s.apply(lambda x: x.month).to_numpy()
    df['daysleftofmonth'] = index_s.apply(lambda x: x.days_in_month - x.day).to_numpy()

    # 本月第几个交易日, 本月还剩几个交易日
    groups = trade_date_series.groupby(trade_date_series.apply(lambda x: datetime.datetime(x.year, x.month, 1)))

    def get_td_of_month(dt):
        first_day_of_month = datetime.datetime(dt.year, dt.month, 1)
        month_s = groups.get_group(first_day_of_month)
        td_of_month = (month_s <= dt).sum()
        td_left_of_month = (month_s > dt).sum()
        return month_s.shape[0], td_of_month, td_left_of_month

    result_arr = index_s.apply(get_td_of_month).to_numpy()
    df['td_of_month'] = [_[0] for _ in result_arr]
    df['td_pass_of_month'] = [_[1] for _ in result_arr]
    df['td_left_of_month'] = [_[2] for _ in result_arr]

    # 本周第几个交易日, 本周还剩几个交易日
    groups = trade_date_series.groupby(trade_date_series.apply(lambda x: x.year * 100 + x.weekofyear))

    def get_td_of_week(dt):
        name = dt.year * 100 + dt.weekofyear
        week_s = groups.get_group(name)
        td_pass_of_week = (week_s <= dt).sum()
        td_left_of_week = (week_s > dt).sum()
        return week_s.shape[0], td_pass_of_week, td_left_of_week

    result_arr = index_s.apply(get_td_of_week).to_numpy()
    df['td_of_week'] = [_[0] for _ in result_arr]
    df['td_pass_of_week'] = [_[1] for _ in result_arr]
    df['td_left_of_week'] = [_[2] for _ in result_arr]

    # 距离下一次放假交易日数（超过2天以上的休息日）
    # 计算距离下一个交易日的日期
    days_2_next_trade_date_s = (index_s.shift(-1) - index_s).fillna(pd.Timedelta(days=0))
    # 倒序循环，计算距离下一次放假的日期
    days_count, trade_date_count, result_arr = np.nan, np.nan, []
    for _, delta in days_2_next_trade_date_s.sort_index(ascending=False).items():
        days = delta.days
        if days > 3:
            trade_date_count, days_count = 0, 0
        elif pd.isna(days_count) >= 0 and days >= 1:
            trade_date_count += 1
            days_count += days
        else:
            trade_date_count, days_count = np.nan, np.nan

        result_arr.append([days_count, trade_date_count])

    result_arr.reverse()
    df['days_2_vacation'] = [_[0] for _ in result_arr]
    df['td_2_vacation'] = [_[1] for _ in result_arr]

    return df


def add_factor_of_delivery_date(df, delivery_date_series):
    index_s = pd.Series(df.index)

    # 当期合约距离交割日天数
    result_arr = []
    for _, trade_date in index_s.items():
        next_2_date = delivery_date_series[delivery_date_series > trade_date].head(2)
        if next_2_date.shape[0] < 2:
            day_2_first_del_date, day_2_second_del_date = np.nan, np.nan
        else:
            first_del_date, secend_del_date = next_2_date[0], next_2_date[1]
            day_2_first_del_date = (first_del_date - trade_date).days
            day_2_second_del_date = (secend_del_date - trade_date).days

        result_arr.append([day_2_first_del_date, day_2_second_del_date])

    df['days_2_first_del_date'] = [_[0] for _ in result_arr]
    df['days_2_second_del_date'] = [_[1] for _ in result_arr]
    return df


def add_factor_of_price(df: pd.DataFrame, close_key='close'):
    # 均线因子
    close_s = df[close_key]
    df[f'rr'] = close_s.to_returns()
    for n in [5, 10, 15, 20, 30, 60]:
        df[f'ma{n}'] = close_s.rolling(n).mean()
    # EMA
    for n in [12, 26, 60]:
        df[f'ema{n}'] = close_s.ewm(span=n).mean()
    # 波动率因子
    expanding = close_s.expanding(5)
    df[f'volatility_all'] = expanding.std() / expanding.mean()
    for n in [20, 60, 120, 240]:
        df[f'volatility{n}'] = close_s.rolling(n).std() / close_s.rolling(n).mean()

    # 收益率方差
    rr = close_s.to_returns()
    for n in [20, 60]:
        df[f'rr_std{n}'] = rr.rolling(n).std()

    df[f'ema{n}'] = close_s.ewm(span=n).mean()
    return df


def get_factor(df: pd.DataFrame, close_key='close', vol_key='volume', trade_date_series=None, delivery_date_series=None, dropna=True) -> pd.DataFrame:
    """
    在当期时间序列数据基础上增加相关因子
    目前已经支持的因子包括量价因子、时间序列因子、交割日期因子
    :param df: 时间序列数据索引为日期
    :param close_key:
    :param vol_key:
    :param trade_date_series: 交易日序列
    :param delivery_date_series:交割日序列
    :param dropna: 是否 dropna
    :return:
    """
    ret_df = df.copy()

    # 增加量价因子
    if close_key is not None:
        ret_df = add_factor_of_price(ret_df, close_key)

    # 交易日相关因子
    if trade_date_series is not None:
        ret_df = add_factor_of_trade_date(ret_df, trade_date_series)

    # 交割日相关因子
    if delivery_date_series is not None:
        ret_df = add_factor_of_delivery_date(ret_df, delivery_date_series)

    if dropna:
        ret_df.dropna(inplace=True)
    return ret_df


def _test_get_factor():
    from ibats_common.example.data import load_data, get_trade_date_series
    instrument_type = 'RB'
    file_name = f"{instrument_type}.csv"
    df = load_data(file_name).set_index('trade_date').drop('instrument_type', axis=1)
    df.index = pd.DatetimeIndex(df.index)
    factor = get_factor(df,
                        trade_date_series=get_trade_date_series(),
                        delivery_date_series=get_delivery_date_series(instrument_type))
    logger.info("\n%s\t%s \n%s\t%s", df.shape, list(df.columns), factor.shape, list(factor.columns))
    print("\n%s\t%s \n%s\t%s" % (df.shape, list(df.columns), factor.shape, list(factor.columns)))


if __name__ == '__main__':
    import ibats_common.config
    _test_get_factor()



#! /usr/bin/env python
# -*- coding:utf-8 -*-
"""
@author  : MG
@Time    : 2019/3/8 12:23
@File    : orm_stg_run_status_detail_test.py
@contact : mmmaaaggg@163.com
@desc    : 
"""
from ibats_common.backend.orm import *
from datetime import date, datetime, time
import unittest
from ibats_common.test.orm_init_test import InitTest

from ibats_common.common import ExchangeName


class TradeDetailTest(unittest.TestCase):  # 继承unittest.TestCase
    def tearDown(self):
        pass

    def setUp(self):
        # 每个测试用例执行之前做操作
        pass

    @classmethod
    def tearDownClass(cls):
        # 必须使用 @ classmethod装饰器, 所有test运行完后运行一次
        pass

    @classmethod
    def setUpClass(cls):
        # from ibats_common.config import update_db_config, ConfigBase
        # update_db_config({
        #     ConfigBase.DB_SCHEMA_IBATS: 'mysql://mg:Dcba1234@localhost/' + ConfigBase.DB_SCHEMA_IBATS,
        # })
        # 必须使用@classmethod 装饰器,所有test运行前运行一次
        init()
        global engine_ibats
        engine_ibats = engines.engine_ibats

    def test_create_by_order_detail(self):
        order = InitTest.add_order()
        trade = TradeDetail.create_by_order_detail(order)
        self.assertIsInstance(trade, TradeDetail)
        self.assertGreater(trade.commission, 0)
        self.assertGreater(trade.margin, 0)
        self.assertGreater(trade.multiple, 0)
        self.assertGreater(trade.margin_ratio, 0)
        self.assertEqual(trade.trade_vol, order.order_vol)
        self.assertEqual(trade.trade_price, order.order_price)
        self.assertEqual(trade.direction, order.direction)
        self.assertEqual(trade.action, order.action)


if __name__ == '__main__':
    unittest.main()  # 运行所有的测试用例



# -*- coding: utf-8 -*-
"""
Created on 2017/10/3
@author: MG
"""
import logging
from ibats_common.common import RunMode, ExchangeName
from ibats_common.trade import trader_agent, BacktestTraderAgentBase, FixPercentBacktestTraderAgentBase
from enum import Enum

logger = logging.getLogger()


class OrderType(Enum):
    """
    buy-market：市价买, sell-market：市价卖, buy-limit：限价买, sell-limit：限价卖, buy-ioc：IOC买单, sell-ioc：IOC卖单
    """
    buy_market = 'buy-market'
    sell_market = 'sell-market'
    buy_limit = 'buy-limit'
    sell_limit = 'sell-limit'
    buy_ioc = 'buy-ioc'
    sell_ioc = 'sell-ioc'


@trader_agent(RunMode.Backtest, ExchangeName.LocalFile, is_default=False)
class BacktestTraderAgent(BacktestTraderAgentBase):
    """
    供调用模拟交易接口使用
    """

    def __init__(self, stg_run_id, **kwargs):
        super().__init__(stg_run_id, **kwargs)


@trader_agent(RunMode.Backtest_FixPercent, ExchangeName.LocalFile, is_default=False)
class FixPercentBacktestTraderAgent(FixPercentBacktestTraderAgentBase):
    """
    供调用模拟交易接口使用
    """

    def __init__(self, stg_run_id, **kwargs):
        super().__init__(stg_run_id, **kwargs)


if __name__ == "__main__":
    import time

    # 测试交易 下单接口及撤单接口
    # symbol, vol, price = 'ocnusdt', 1, 0.00004611  # OCN/USDT
    symbol, vol, price = 'eosusdt', 1.0251, 4.1234  # OCN/USDT

    td = BacktestTraderAgent(stg_run_id=1, run_mode_params={})
    td.open_long(symbol=symbol, price=price, vol=vol)
    order_dic_list = td.get_order(symbol=symbol)
    print('after open_long', order_dic_list)
    assert len(order_dic_list) == 1
    td.cancel_order(instrument_id=symbol)
    time.sleep(1)
    order_dic_list = td.get_order(symbol=symbol)
    print('after cancel', order_dic_list)
    assert len(order_dic_list) == 0



# -*- coding: utf-8 -*-
"""
Created on 2017/10/9
@author: MG
"""
import logging
import warnings
from collections import defaultdict
from datetime import datetime, timedelta

import numpy as np
import pandas as pd
from ibats_utils.db import with_db_session
from ibats_utils.mess import str_2_date, pd_timedelta_2_timedelta, datetime_2_str, date_time_2_str, try_2_datetime
from sqlalchemy import Column, Integer, String, DateTime, SmallInteger, Date, Time
from sqlalchemy.dialects.mysql import DOUBLE, TINYINT
from sqlalchemy.ext.declarative import declarative_base

from ibats_common.backend import engines
from ibats_common.common import Action, Direction, CalcMode, ExchangeName, RunMode
from ibats_common.common import PositionDateType
from ibats_common.config import config

engine_ibats = engines.engine_ibats
BaseModel = declarative_base()
_key_idx = defaultdict(lambda: defaultdict(int))
MAX_RATE = 9.99  # 相当于 999%


def idx_generator(key1, key2):
    idx = _key_idx[key1][key2]
    idx += 1
    _key_idx[key1][key2] = idx
    return idx


class StgRunInfo(BaseModel):
    """策略运行信息"""

    __tablename__ = 'stg_run_info'
    stg_run_id = Column(Integer, autoincrement=True, primary_key=True)
    stg_name = Column(String(100))
    stg_module = Column(String(500))
    dt_from = Column(DateTime())
    dt_to = Column(DateTime())
    run_mode = Column(SmallInteger)
    stg_params = Column(String(5000))
    strategy_handler_param = Column(String(1000))
    md_agent_params_list = Column(String(5000))
    trade_agent_params_list = Column(String(5000))


class OrderDetail(BaseModel):
    """订单信息"""

    __tablename__ = 'order_detail'
    stg_run_id = Column(Integer, primary_key=True)
    order_idx = Column(Integer, primary_key=True)
    trade_agent_key = Column(String(40))
    order_dt = Column(DateTime)
    order_date = Column(Date)  # 对应行情数据中 ActionDate
    order_time = Column(Time)  # 对应行情数据中 ActionTime
    order_millisec = Column(Integer, default=0, server_default='0')  # 对应行情数据中 ActionMillisec
    direction = Column(TINYINT)  # -1：空；1：多
    action = Column(Integer)  # 0：开仓：1-6：平仓，具体参见 ibats_common.common.Action 对象
    symbol = Column(String(30))
    order_price = Column(DOUBLE)
    order_vol = Column(DOUBLE)  # 订单量
    calc_mode = Column(TINYINT)  # 计算模式：0 普通模式，1 保证金模式

    def __init__(self, stg_run_id=None, trade_agent_key=None, order_dt=None, order_date=None, order_time=None,
                 order_millisec=0, direction=None, action=None, symbol=None, order_price=0, order_vol=0,
                 calc_mode: (CalcMode, int) = 0):
        self.stg_run_id = stg_run_id
        self.order_idx = None if stg_run_id is None else idx_generator(stg_run_id, OrderDetail)
        self.trade_agent_key = trade_agent_key.name if isinstance(trade_agent_key, ExchangeName) else trade_agent_key
        self.order_dt = date_time_2_str(order_date, order_time) if order_dt is None else order_dt
        self.order_date = order_date
        self.order_time = order_time
        self.order_millisec = order_millisec
        self.direction = direction.value if isinstance(direction, Direction) else direction
        self.action = action.value if isinstance(action, Action) else action
        self.symbol = symbol
        self.order_price = order_price
        self.order_vol = order_vol
        self.calc_mode = calc_mode.value if isinstance(calc_mode, CalcMode) else calc_mode

    def __repr__(self):
        return f"<OrderDetail(stg_run_id='{self.stg_run_id}' idx='{self.order_idx}', " \
               f"trade_agent_key={self.trade_agent_key}, direction='{Direction(self.direction)}', " \
               f"action='{Action(self.action)}', symbol='{self.symbol}', order_price='{self.order_price}', " \
               f"order_vol='{self.order_vol}')>"

    @staticmethod
    def remove(stg_run_id: int):
        """
        仅作为调试工具使用，删除指定 stg_run_id 相关的 order_detail
        :param stg_run_id: 
        :return: 
        """
        with with_db_session(engine_ibats) as session:
            # session.execute('DELETE FROM order_detail WHERE stg_run_id=:stg_run_id',
            #                 {'stg_run_id': stg_run_id})
            session.query(OrderDetail).filter(OrderDetail.stg_run_id == stg_run_id).delete()
            session.commit()


class TradeDetail(BaseModel):
    """记录成交信息"""
    __tablename__ = 'trade_detail'
    stg_run_id = Column(Integer, primary_key=True)
    trade_idx = Column(Integer, primary_key=True)  # , comment="成交id"
    trade_agent_key = Column(String(40))
    order_idx = Column(Integer)  # , comment="对应订单id"
    order_price = Column(DOUBLE)  # , comment="原订单价格"
    order_vol = Column(DOUBLE)  # 订单量 , comment="原订单数量"
    trade_dt = Column(DateTime)  # 对应行情数据中 ActionDate
    trade_date = Column(Date)  # 对应行情数据中 ActionDate
    trade_time = Column(Time)  # 对应行情数据中 ActionTime
    trade_millisec = Column(Integer)  # 对应行情数据中 ActionMillisec
    direction = Column(TINYINT)  # -1：空；1：多
    action = Column(Integer)  # 0：开仓：1-6：平仓，具体参见 ibats_common.common.Action 对象
    symbol = Column(String(30))
    trade_price = Column(DOUBLE)  # , comment="成交价格"
    trade_vol = Column(DOUBLE)  # 订单量 , comment="成交数量"
    margin = Column(DOUBLE, server_default='0')  # 保证金 , comment="占用保证金"
    commission = Column(DOUBLE, server_default='0')  # 佣金、手续费 , comment="佣金、手续费"
    multiple = Column(DOUBLE, server_default='0')  # 合约乘数
    margin_ratio = Column(DOUBLE, server_default='0')  # 保证金比例
    calc_mode = Column(TINYINT)  # 计算模式：0 普通模式，1 保证金模式

    def __init__(self, stg_run_id=None, trade_agent_key=None, order_idx=None, order_price=None, order_vol=None,
                 trade_dt=None, trade_date=None, trade_time=None, trade_millisec=0, direction=None, action=None,
                 symbol=None, trade_price=None, trade_vol=None, margin=None, commission=None, multiple=None,
                 margin_ratio=None, calc_mode: (CalcMode, int) = 0):
        self.stg_run_id = stg_run_id
        self.trade_idx = None if stg_run_id is None else idx_generator(stg_run_id, TradeDetail)
        self.trade_agent_key = trade_agent_key.name if isinstance(trade_agent_key, ExchangeName) else trade_agent_key
        self.order_idx = order_idx
        self.order_price = order_price
        self.order_vol = order_vol
        self.trade_dt = try_2_datetime(date_time_2_str(trade_date, trade_time)) if trade_dt is None else trade_dt
        self.trade_date = trade_date
        self.trade_time = trade_time
        self.trade_millisec = trade_millisec
        self.direction = direction
        self.action = action
        self.symbol = symbol
        self.trade_price = trade_price
        self.trade_vol = trade_vol
        self.margin = margin
        self.commission = commission
        self.multiple = multiple
        self.margin_ratio = margin_ratio
        self.calc_mode = calc_mode.value if isinstance(calc_mode, CalcMode) else calc_mode

    def set_trade_time(self, value):
        if isinstance(value, pd.Timedelta):
            # print(value, 'parse to timedelta')
            self.trade_time = timedelta(seconds=value.seconds)
        else:
            self.trade_time = value

    @staticmethod
    def remove(stg_run_id: int):
        """
        仅作为调试工具使用，删除指定 stg_run_id 相关的 trade_detail
        :param stg_run_id: 
        :return: 
        """
        with with_db_session(engine_ibats) as session:
            # session.execute('DELETE FROM trade_detail WHERE stg_run_id=:stg_run_id',
            #                 {'stg_run_id': stg_run_id})
            # session.query(PosStatusInfo).filter(PosStatusInfo.stg_run_id.in_([3, 4])).delete(False)  # 仅供实例
            session.query(TradeDetail).filter(TradeDetail.stg_run_id == stg_run_id).delete()
            session.commit()

    @staticmethod
    def create_by_order_detail(order_detail: OrderDetail):
        symbol = order_detail.symbol
        order_price, order_vol = order_detail.order_price, order_detail.order_vol
        # stg_run_id = order_detail.stg_run_id

        # TODO: 增加滑点、成交比例、费率、保证金比例等参数，该参数将和在回测参数中进行设置
        # instrument_info = Config.instrument_info_dic[symbol]
        # multiple = instrument_info['VolumeMultiple']
        # margin_ratio = instrument_info['LongMarginRatio']
        multiple, margin_ratio, commission_rate = 1, 1, 0.0005
        margin = order_vol * order_price * multiple * margin_ratio
        commission = order_vol * order_price * multiple * commission_rate
        detail = TradeDetail(stg_run_id=order_detail.stg_run_id,
                             trade_agent_key=order_detail.trade_agent_key,
                             order_idx=order_detail.order_idx,
                             trade_date=order_detail.order_date,
                             trade_time=order_detail.order_time,
                             trade_millisec=order_detail.order_millisec,
                             direction=order_detail.direction,
                             action=order_detail.action,
                             symbol=symbol,
                             order_price=order_price,
                             order_vol=order_vol,
                             trade_price=order_price,
                             trade_vol=order_vol,
                             margin=margin,
                             commission=commission,
                             multiple=multiple,
                             margin_ratio=margin_ratio,
                             calc_mode=order_detail.calc_mode,
                             )
        if config.ORM_UPDATE_OR_INSERT_PER_ACTION:
            with with_db_session(engine_ibats, expire_on_commit=False) as session:
                session.add(detail)
                session.commit()
        return detail


class PosStatusDetail(BaseModel):
    """
    持仓状态数据
    当持仓状态从有仓位到清仓时（position>0 --> position==0），计算清仓前的浮动收益，并设置到 floating_pl 字段最为当前状态的浮动收益
    在调用 create_by_self 时，则需要处理一下，当 position==0 时，floating_pl 直接设置为 0，避免引起后续计算上的混淆
    2018-11-02 当仓位多空切换时（1根K线内多头反手转空头），则浮动收益继续保持之前数字
    """
    __tablename__ = 'pos_status_detail'
    stg_run_id = Column(Integer, primary_key=True)  # 对应回测了策略 StgRunID 此数据与 AccSumID 对应数据相同
    pos_status_detail_idx = Column(Integer, primary_key=True)
    trade_agent_key = Column(String(40))
    trade_idx = Column(Integer)  # , comment="最新的成交id"
    trade_dt = Column(DateTime)  # 每个订单变化生成一条记录 此数据与 AccSumID 对应数据相同
    trade_date = Column(Date)  # 对应行情数据中 ActionDate
    trade_time = Column(Time)  # 对应行情数据中 ActionTime
    trade_millisec = Column(SmallInteger)  # 对应行情数据中 ActionMillisec
    direction = Column(TINYINT)
    symbol = Column(String(30))
    position = Column(DOUBLE, default=0.0)
    position_chg = Column(DOUBLE, default=0.0)
    position_value = Column(DOUBLE, default=0.0)  # 持仓投资品种的总市值 position * trade_price * multiple
    avg_price = Column(DOUBLE, default=0.0)  # 所持投资品种上一交易日所有交易的加权平均价
    cur_price = Column(DOUBLE, default=0.0)
    # 持仓收益，对于普通账户：
    # (trade_price - avg_price_last) * position * int(pos_status_detail.direction) - commission
    # 对于保证金交易：市值增量 - 保证金占比增量
    # (市场价 - 成本价市值) * 仓位 * 方向 -
    floating_pl = Column(DOUBLE, default=0.0)
    # floating_pl_rate 与保证金比例 以及 乘数 变化没有关系
    # (trade_price - avg_price) / avg_price * multiple * int(pos_status_detail.direction)
    floating_pl_rate = Column(DOUBLE, default=0.0)
    floating_pl_chg = Column(DOUBLE, default=0.0)
    floating_pl_cum = Column(DOUBLE, default=0.0)
    # 记录当前状态与前一状态之间净现金流变化情况
    # 例如：
    # 加仓，将导致净现金流为负，减仓则净现金流为正
    # 在保证金交易的情况下，价格波动引起的保证金占用变化，也会使得净现金流产生变化
    cashflow = Column(DOUBLE, default=0.0)
    # 记录每日现金流变化情况
    cashflow_daily = Column(DOUBLE, default=0.0)
    # 累计现金流，整个执行周期内的累计现金净流入量。
    # 该字段将用于与 TradeAgentStatusDetail.cash_init 相加，计算 cash_available 当前可用现金
    cashflow_cum = Column(DOUBLE, default=0.0)
    rr = Column(DOUBLE, default=0.0)  # floating_pl_cum / margin 如果是清仓，则使用前一时刻 margin
    margin = Column(DOUBLE, default=0.0)
    margin_chg = Column(DOUBLE, default=0.0)
    position_date_type = Column(TINYINT, default=0)
    commission = Column(DOUBLE, default=0)  # 当前bar上如果存在交易，则相应的费用记录在此，其他情况均为0
    commission_tot = Column(DOUBLE, default=0)  # 累计费用
    multiple = Column(DOUBLE, server_default='0')  # 合约乘数
    margin_ratio = Column(DOUBLE, server_default='0')  # 保证金比例
    calc_mode = Column(TINYINT)  # 计算模式：0 普通模式，1 保证金模式
    logger = logging.getLogger(f'<Table:{__tablename__}>')

    def __repr__(self):
        return f"<PosStatusDetail(id='{self.pos_status_detail_idx}', trade_agent_key={self.trade_agent_key}, " \
               f"trade_dt='{datetime_2_str(self.trade_dt)}', trade_idx='{self.trade_idx}', symbol='{self.symbol}', " \
               f"direction='{self.direction}', position='{self.position}', avg_price='{self.avg_price}', " \
               f"floating_pl='{self.floating_pl}', floating_pl_chg='{self.floating_pl_chg}', " \
               f"floating_pl_cum='{self.floating_pl_cum}', cashflow='{self.cashflow}', " \
               f"cashflow_daily='{self.cashflow_daily}', cashflow_cum='{self.cashflow_cum}')>"

    def __init__(self, stg_run_id=None, trade_agent_key=None, trade_idx=None, trade_dt=None, trade_date=None,
                 trade_time=None, trade_millisec=None, direction=None, symbol=None, position=None, position_chg=0.0,
                 avg_price=None, cur_price=None, floating_pl=0.0, floating_pl_rate=0.0, floating_pl_chg=0.0,
                 floating_pl_cum=0.0, cashflow=0.0, cashflow_daily=0.0, cashflow_cum=0.0, rr=0.0, margin=0.0,
                 margin_chg=0.0, position_date_type=PositionDateType.Today.value, commission=0.0, commission_tot=0.0,
                 multiple=0, margin_ratio=0.0, calc_mode: (int, CalcMode) = CalcMode.Normal.value):
        self.stg_run_id = stg_run_id
        self.pos_status_detail_idx = None if stg_run_id is None else idx_generator(stg_run_id, PosStatusDetail)
        self.trade_agent_key = trade_agent_key.name if isinstance(trade_agent_key, ExchangeName) else trade_agent_key
        self.trade_idx = trade_idx
        self.trade_dt = date_time_2_str(trade_date, trade_time) if trade_dt is None else trade_dt
        self.trade_date = trade_date
        self.trade_time = trade_time
        self.trade_millisec = trade_millisec
        self.direction = direction
        self.symbol = symbol
        self.position = position
        self.position_chg = position_chg
        self.position_value = position * cur_price
        self.avg_price = avg_price
        self.cur_price = cur_price
        self.floating_pl = floating_pl
        self.floating_pl_rate = floating_pl_rate
        self.floating_pl_chg = floating_pl_chg
        self.floating_pl_cum = floating_pl_cum
        self.cashflow = cashflow
        self.cashflow_daily = cashflow_daily
        self.cashflow_cum = cashflow_cum
        self.rr = rr
        self.margin = margin
        self.margin_chg = margin_chg
        self.position_date_type = position_date_type
        self.commission = commission
        self.commission_tot = commission_tot
        self.multiple = multiple
        self.margin_ratio = margin_ratio
        self.calc_mode = calc_mode.value if isinstance(calc_mode, CalcMode) else calc_mode
        self.last_status = None  # 记录上一个状态实例
        self.last_date_status = None  # 记录上一日最后一个状态实例

    @staticmethod
    def create_by_trade_detail(trade_detail: TradeDetail):
        direction, action, instrument_id = trade_detail.direction, trade_detail.action, trade_detail.symbol
        if action == int(Action.Close):
            raise ValueError('trade_detail.action 不能为 close')
        trade_vol = trade_detail.trade_vol
        trade_price = trade_detail.trade_price
        commission = trade_detail.commission
        tot_value = trade_vol * trade_price
        margin = trade_detail.margin
        # tot_cost = tot_value + commission
        # avg_price = tot_cost / trade_vol
        # 2019-05-19 bug fix on avg_price
        avg_price = (tot_value + commission * int(direction)) / trade_vol
        floating_pl = -commission
        floating_pl_rate = floating_pl / margin
        cashflow = -margin - commission
        detail = PosStatusDetail(stg_run_id=trade_detail.stg_run_id,
                                 trade_agent_key=trade_detail.trade_agent_key,
                                 trade_idx=trade_detail.trade_idx,
                                 trade_dt=trade_detail.trade_dt,
                                 trade_date=trade_detail.trade_date,
                                 trade_time=trade_detail.trade_time,
                                 trade_millisec=trade_detail.trade_millisec,
                                 direction=trade_detail.direction,
                                 symbol=trade_detail.symbol,
                                 position=trade_vol,
                                 position_chg=trade_vol,
                                 avg_price=avg_price,
                                 cur_price=trade_price,
                                 margin=margin,
                                 margin_chg=margin,
                                 floating_pl=floating_pl,
                                 floating_pl_rate=floating_pl_rate,
                                 floating_pl_chg=floating_pl,
                                 floating_pl_cum=floating_pl,
                                 cashflow=cashflow,
                                 cashflow_daily=cashflow,
                                 cashflow_cum=cashflow,
                                 rr=floating_pl_rate,
                                 commission=commission,
                                 commission_tot=commission,
                                 position_date_type=PositionDateType.Today.value,
                                 multiple=trade_detail.multiple,
                                 margin_ratio=trade_detail.margin_ratio,
                                 calc_mode=trade_detail.calc_mode,
                                 )
        if config.ORM_UPDATE_OR_INSERT_PER_ACTION:
            # 更新最新持仓纪录
            with with_db_session(engine_ibats, expire_on_commit=False) as session:
                session.add(detail)
                session.commit()
        return detail

    def update_by_trade_detail(self, trade_detail: TradeDetail):
        """
        创建新的对象，根据 trade_detail 更新相关信息
        :param trade_detail:
        :return: 
        """
        # 复制前一个持仓状态
        detail = self.create_by_self()
        direction, action, symbol = trade_detail.direction, trade_detail.action, trade_detail.symbol
        trade_price, trade_vol, trade_idx = trade_detail.trade_price, trade_detail.trade_vol, trade_detail.trade_idx
        commission = trade_detail.commission

        # 获取合约信息
        # instrument_info = Config.instrument_info_dic[symbol]
        # multiple = instrument_info['VolumeMultiple']
        # margin_ratio = instrument_info['LongMarginRatio']
        multiple, margin_ratio = 1, 1

        # 计算
        # 仓位 position、position_chg 、
        # 方向 direction、
        # 平均价格 avg_price、
        # 浮动收益 floating_pl floating_pl_rate
        pos_direction_last = self.direction
        position_last = self.position
        avg_price_last = self.avg_price
        margin_last = self.margin

        if position_last == 0:
            # 如果前一状态仓位为 0 则本次方向与当前订单方向相同
            detail.direction = trade_detail.direction

        direction_int = int(detail.direction)
        if self.calc_mode == CalcMode.Normal.value:
            # 普通模式：非保证金交易模式
            # 普通模式 默认 margin_rate, multiple 均为 1
            if pos_direction_last == direction or position_last == 0:
                if action == Action.Open:
                    # 方向相同：开仓 or 加仓；
                    position_cur = position_last + trade_vol
                    position_value = position_cur * trade_price
                    detail.position_value = position_value
                    detail.position_chg = trade_vol
                    detail.position = position_cur
                    avg_price = (position_last * avg_price_last + trade_price * trade_vol +
                                 commission * direction_int) / position_cur
                    detail.avg_price = avg_price
                    # 计算浮动收益 floating_pl floating_pl_rate
                    detail.floating_pl = (trade_price - avg_price) * position_cur * direction_int
                    detail.floating_pl_rate = (trade_price - avg_price) / avg_price * direction_int
                else:
                    # 方向相反：清仓 or 减仓；
                    detail.position_chg = - trade_vol
                    if trade_vol > position_last:
                        raise ValueError("当前持仓%d，平仓%d，错误" % (position_last, trade_vol))
                    elif trade_vol == position_last:
                        # 清仓
                        position_cur = 0
                        position_value = position_cur * trade_price
                        detail.position_value = position_value
                        detail.avg_price = avg_price = 0
                        detail.position = position_cur
                        # 计算浮动收益 floating_pl floating_pl_rate
                        # 与其他地方计算公式的区别在于 position_curr == 0 因此使用 position_last
                        detail.floating_pl = (trade_price - avg_price_last) * position_last * direction_int - commission
                        detail.floating_pl_rate = (
                                ((trade_price - avg_price_last) * direction_int - commission / position_last)
                                / avg_price_last
                        ) if avg_price_last > 0.001 else MAX_RATE

                    else:
                        # 减仓
                        position_cur = position_last - trade_vol
                        position_value = position_cur * trade_price
                        detail.position_value = position_value
                        avg_price = (position_last * avg_price_last - trade_price * trade_vol + commission
                                     ) / position_cur
                        detail.avg_price = avg_price
                        detail.position = position_cur
                        # 计算浮动收益 floating_pl floating_pl_rate
                        detail.floating_pl = (trade_price - avg_price) * position_cur * direction_int
                        detail.floating_pl_rate = (trade_price - avg_price) / avg_price * direction_int

            else:
                # 方向相反
                raise ValueError("当前仓位：%s %d手，目标操作：%s %d手，请先平仓在开仓" % (
                    "多头" if pos_direction_last == Direction.Long else "空头", position_last,
                    "多头" if direction == Direction.Long else "空头", trade_vol,
                ))

            # 设置其他属性 floating_pl_chg、floating_pl_cum、cur_price、trade_dt、trade_date、trade_time、trade_millisec
            # 2019-05-19 当 position_last == 0 时，代表本次交易为重新开仓
            # 因此，floating_pl_chg = floating_pl，而非与上一状态的差
            detail.floating_pl_chg = (detail.floating_pl - self.floating_pl
                                      ) if position_last != 0 else detail.floating_pl
            detail.floating_pl_cum += detail.floating_pl_chg

            # 计算 position_value、margin、margin_chg
            # cur_price = pos_status_detail.cur_price
            detail.margin = position_cur * trade_price
            # 如果前一状态仓位为 0,  且不是多空切换的情况，则保留上一状态的浮动收益
            if self.position == 0 and self.trade_dt != trade_detail.trade_dt:
                # 新建仓情况下，margin_last 为当前 margin
                # 该变量用于后续计算 rr 使用
                margin_last = detail.margin
                detail.margin_chg = margin_chg = detail.margin
            else:
                margin_last_pos_cur_price = self.position * trade_price
                detail.margin_chg = margin_chg = detail.margin - margin_last_pos_cur_price

            # 计算 cashflow_daily, cashflow_cum, commission, commission_tot, rr, position_date_type
            # 本次现金流
            detail.cashflow = cashflow = - margin_chg - commission
            # 每日现金流
            if self.trade_date != trade_detail.trade_date:
                detail.cashflow_daily = cashflow
            else:
                detail.cashflow_daily += cashflow

            # 累计现金流
            detail.cashflow_cum += cashflow
            detail.commission = commission
            detail.commission_tot += commission
            detail.rr = detail.floating_pl_cum / (
                detail.margin if detail.margin > 0 else margin_last)
            detail.position_date_type = PositionDateType.Today.value

        elif self.calc_mode == CalcMode.Margin.value:
            # 保证金交易模式
            if pos_direction_last == direction or position_last == 0:
                if action == Action.Open:
                    # 方向相同：开仓 or 加仓；
                    position_cur = position_last + trade_vol
                    position_value = position_cur * trade_price * multiple
                    detail.position_value = position_value
                    detail.position_chg = trade_vol
                    detail.position = position_cur
                    avg_price = (position_last * avg_price_last * multiple + trade_price * trade_vol * multiple
                                 + commission * direction_int) / position_cur
                    detail.avg_price = avg_price
                    # 计算 margin、margin_chg
                    margin = position_value * margin_ratio
                    detail.margin = margin
                    margin_chg = margin - margin_last
                    detail.margin_chg = margin_chg
                    # 计算浮动收益 floating_pl floating_pl_rate
                    detail.floating_pl = (trade_price - avg_price) * position_cur * multiple * direction_int
                    detail.floating_pl_rate = (trade_price - avg_price) / avg_price * direction_int
                else:
                    # 方向相反：清仓 or 减仓；
                    detail.position_chg = - trade_vol
                    if trade_vol > position_last:
                        raise ValueError("当前持仓%d，平仓%d，错误" % (position_last, trade_vol))
                    elif trade_vol == position_last:
                        # 清仓
                        position_cur = 0
                        position_value = position_cur * trade_price * multiple
                        detail.position_value = position_value
                        detail.avg_price = avg_price = 0
                        detail.position = position_cur
                        # 计算 margin、margin_chg
                        margin_last_pos_cur_price = position_last * trade_price * multiple * margin_ratio
                        margin = position_value * margin_ratio
                        detail.margin = margin
                        margin_chg = margin - margin_last_pos_cur_price
                        detail.margin_chg = margin_chg
                        # 计算浮动收益 floating_pl floating_pl_rate
                        # 与其他地方计算公式的区别在于 position_curr == 0 因此使用 position_last
                        detail.floating_pl = (trade_price - avg_price_last
                                              ) * position_last * multiple * direction_int - commission
                        detail.floating_pl_rate = (
                                ((trade_price - avg_price_last) * direction_int - commission / position_last)
                                / avg_price_last
                        )if avg_price_last > 0.001 else MAX_RATE

                    else:
                        # 减仓
                        position_cur = position_last - trade_vol
                        position_value = position_cur * trade_price * multiple
                        detail.position_value = position_value
                        avg_price = (position_last * avg_price_last * multiple - trade_price * trade_vol * multiple
                                     + commission) / position_cur
                        detail.avg_price = avg_price
                        detail.position = position_cur
                        # 计算 margin、margin_chg
                        margin = position_value * margin_ratio
                        detail.margin = margin
                        margin_chg = margin - margin_last
                        detail.margin_chg = margin_chg
                        # 计算浮动收益 floating_pl floating_pl_rate
                        detail.floating_pl = (trade_price - avg_price) * position_cur * multiple * direction_int
                        detail.floating_pl_rate = (trade_price - avg_price) / avg_price * direction_int

            else:
                # 方向相反
                raise ValueError("当前仓位：%s %d手，目标操作：%s %d手，请先平仓在开仓" % (
                    "多头" if pos_direction_last == Direction.Long else "空头", position_last,
                    "多头" if direction == Direction.Long else "空头", trade_vol,
                ))

            # 设置其他属性 floating_pl_chg、floating_pl_cum、cur_price、trade_dt、trade_date、trade_time、trade_millisec
            # 2019-05-19 当 position_last == 0 时，代表本次交易为重新开仓
            # 因此，floating_pl_chg = floating_pl，而非与上一状态的差
            detail.floating_pl_chg = (detail.floating_pl - self.floating_pl
                                      ) if position_last != 0 else detail.floating_pl
            detail.floating_pl_cum += detail.floating_pl_chg

            # 计算 cashflow_daily、commission、commission_tot、rr、position_date_type
            # 本次现金流
            detail.cashflow = cashflow = - margin_chg - commission
            # 每日现金流
            if self.trade_date != trade_detail.trade_date:
                detail.cashflow_daily = cashflow
            else:
                detail.cashflow_daily += cashflow

            # 累计现金流
            detail.cashflow_cum += cashflow
            detail.commission += commission
            detail.commission_tot += commission
            detail.rr = detail.floating_pl_cum / (
                detail.margin if detail.margin > 0 else margin_last)
            detail.position_date_type = PositionDateType.Today.value

        else:
            ValueError('calc_mode 不是有效的值 %s', self.calc_mode)

        detail.cur_price = trade_price
        detail.trade_dt = trade_detail.trade_dt
        detail.trade_date = trade_detail.trade_date
        detail.trade_time = trade_detail.trade_time
        detail.trade_millisec = trade_detail.trade_millisec

        # self.logger.debug("%s", pos_status_detail)

        if config.ORM_UPDATE_OR_INSERT_PER_ACTION:
            # 更新最新持仓纪录
            with with_db_session(engine_ibats, expire_on_commit=False) as session:
                session.add(detail)
                session.commit()
        return detail

    def update_by_md(self, trade_price, timestamp_curr: (datetime, pd.Timestamp) = None,
                     md: dict = None,
                     timestamp_key=None, date_key=None, time_key=None, milli_sec_key=None, close_key=None):
        """
        根据行情 更新 pos_status_detail 生成新的对象
        :param trade_price:行情报价
        :param timestamp_curr: timestamp_curr 或 md + key 参数 两组参数二选其一
        :param md: timestamp_curr 或 md + key 参数 两组参数二选其一
        :param timestamp_key: timestamp_curr 或 md + key 参数 两组参数二选其一
        :param date_key: timestamp_curr 或 md + key 参数 两组参数二选其一
        :param time_key: timestamp_curr 或 md + key 参数 两组参数二选其一
        :param milli_sec_key: timestamp_curr 或 md + key 参数 两组参数二选其一
        :param close_key: price 或 md + close_key 参数 两组参数二选其一
        :return:
        """
        # 更新日期、时间
        if timestamp_curr is None:
            timestamp_curr = md[timestamp_key]

        if timestamp_curr is None:
            trade_date = str_2_date(md[date_key]) - timedelta(days=1)
            trade_time = pd_timedelta_2_timedelta(md[time_key])
            trade_dt = datetime_2_str(date_time_2_str(trade_date, trade_time))
            trade_millisec = int(md.setdefault(milli_sec_key, 0)) if milli_sec_key is not None else 0
        else:
            trade_date = timestamp_curr.date()
            trade_time = timestamp_curr.time()
            trade_dt = timestamp_curr
            trade_millisec = int(md.setdefault(milli_sec_key, 0)) if milli_sec_key is not None else 0

        # 设置价格
        if trade_price is None:
            trade_price = md[close_key]

        is_new_day = self.trade_date != trade_date
        detail = self.create_by_self(is_new_day=is_new_day)
        detail.cur_price = trade_price
        detail.trade_date = trade_date
        detail.trade_time = trade_time
        detail.trade_dt = trade_dt
        detail.trade_millisec = trade_millisec

        # 乘数、保证金暂时按1来计算
        # TODO: 增加保证金，乘数参数
        multiple, margin_ratio = 1, 1
        # 没有交易，不产生手续费
        commission = 0
        position_cur, avg_price, direction_int = detail.position, detail.avg_price, int(detail.direction)
        position_value = position_cur * trade_price * multiple
        floating_pl = (trade_price - avg_price) * position_cur * multiple * direction_int
        detail.position_value = position_value
        # 计算 floating_pl
        detail.floating_pl = floating_pl
        # 设置属性 floating_pl_chg、floating_pl_cum
        floating_pl_chg = floating_pl - self.floating_pl
        detail.floating_pl_chg = floating_pl_chg
        detail.floating_pl_cum += floating_pl_chg

        if self.calc_mode == CalcMode.Normal.value:
            # 普通模式：非保证金交易模式
            detail.margin = position_cur * trade_price
            margin_last = self.margin
            detail.margin_chg = margin_chg = detail.margin - margin_last
            # 非保证金模式下，行情变化，不会对现金流产生影响，只会引起浮动收益变化
            detail.cashflow = cashflow = 0
        elif self.calc_mode == CalcMode.Margin.value:
            # 保证金交易模式

            # 计算 position_value、margin、margin_chg
            # cur_price = pos_status_detail.cur_price
            detail.margin = position_cur * trade_price * multiple * margin_ratio
            margin_last = self.margin
            detail.margin_chg = margin_chg = detail.margin - margin_last
            # 计算 cashflow_daily, cashflow_cum, commission, commission_tot, rr, position_date_type
            # 本次现金流
            # 保证金模式下的现金流变化 = 盈利增量 - 保证金增量 - 手续费
            # 盈利增量 = 持仓市值增量 × 方向
            detail.cashflow = cashflow = (position_value - self.position_value
                                          ) * direction_int - margin_chg - commission
            # 每日现金流
            if self.trade_date != trade_date:
                detail.cashflow_daily = cashflow
            else:
                detail.cashflow_daily += cashflow

        else:
            ValueError('calc_mode 不是有效的值 %s', self.calc_mode)

        # 累计现金流
        detail.cashflow_cum += cashflow
        detail.commission = commission
        detail.commission_tot += commission
        # 如果无法计算收益率则暂时记 np.nan
        detail.rr = (detail.floating_pl_cum / detail.margin) if detail.margin > 0 else np.nan

        if config.ORM_UPDATE_OR_INSERT_PER_ACTION:
            # 更新最新持仓纪录
            with with_db_session(engine_ibats, expire_on_commit=False) as session:
                session.add(detail)
                session.commit()

        return detail

    def create_by_self(self, is_new_day=False):
        """
        创建新的对象
        若当前对象持仓为0（position==0），则 浮动收益部分设置为0
        :param is_new_day:
        :return:
        """
        position = self.position
        if is_new_day:
            cashflow_daily = 0
            position_date_type = PositionDateType.History.value
        else:
            cashflow_daily = self.cashflow_daily
            position_date_type = self.position_date_type

        detail = PosStatusDetail(stg_run_id=self.stg_run_id,
                                 trade_agent_key=self.trade_agent_key,
                                 trade_idx=self.trade_idx,
                                 trade_dt=self.trade_dt,
                                 trade_date=self.trade_date,
                                 trade_time=self.trade_time,
                                 trade_millisec=self.trade_millisec,
                                 direction=self.direction,
                                 symbol=self.symbol,
                                 position=position,
                                 avg_price=self.avg_price,
                                 cur_price=self.cur_price,
                                 floating_pl=self.floating_pl if position > 0 else 0,
                                 floating_pl_rate=0.0,
                                 floating_pl_cum=self.floating_pl_cum,
                                 cashflow=0.0,
                                 cashflow_daily=cashflow_daily,
                                 cashflow_cum=self.cashflow_cum,
                                 margin=self.margin,
                                 margin_chg=0,
                                 position_date_type=position_date_type,
                                 commission=0.0,
                                 commission_tot=self.commission_tot,
                                 multiple=self.multiple,
                                 margin_ratio=self.margin_ratio,
                                 calc_mode=self.calc_mode
                                 )
        detail.last_status = self
        if is_new_day:
            detail.last_date_status = self
        else:
            detail.last_date_status = self.last_date_status

        return detail

    @staticmethod
    def remove(stg_run_id: int):
        """
        仅作为调试工具使用，删除指定 stg_run_id 相关的 pos_status_detail
        :param stg_run_id: 
        :return: 
        """
        with with_db_session(engine_ibats) as session:
            # session.execute('DELETE FROM pos_status_detail WHERE stg_run_id=:stg_run_id',
            #                 {'stg_run_id': stg_run_id})
            session.query(PosStatusDetail).filter(PosStatusDetail.stg_run_id == stg_run_id).delete()
            session.commit()
        # PosStatusInfo.query.filter


class TradeAgentStatusDetail(BaseModel):
    """持仓状态数据"""
    __tablename__ = 'trade_agent_status_detail'
    stg_run_id = Column(Integer, primary_key=True)  # 对应回测了策略 StgRunID 此数据与 AccSumID 对应数据相同
    trade_agent_status_detail_idx = Column(Integer, primary_key=True)
    trade_agent_key = Column(String(40))
    trade_dt = Column(DateTime)
    trade_date = Column(Date)  # 对应行情数据中 ActionDate
    trade_time = Column(Time)  # 对应行情数据中 ActionTime
    trade_millisec = Column(Integer)  # 对应行情数据中 ActionMillisec
    # 可用资金, double
    # detail.cash_init + close_profit - curr_margin - commission + (position_value - curr_margin)
    # 对于没有杠杆的产品 position_value == curr_margin 因此 (position_value - curr_margin) == 0
    cash_available_last_day = Column(DOUBLE, default=0.0)
    cash_available = Column(DOUBLE, default=0.0)
    position_value = Column(DOUBLE, default=0.0)
    curr_margin = Column(DOUBLE, default=0.0)  # 当前保证金总额, double
    close_profit = Column(DOUBLE, default=0.0)
    position_profit = Column(DOUBLE, default=0.0)
    floating_pl_cum = Column(DOUBLE, default=0.0)
    commission_tot = Column(DOUBLE, default=0.0)
    cash_init = Column(DOUBLE, default=0.0)
    cash_and_margin = Column(DOUBLE, default=0.0)
    cashflow_daily = Column(DOUBLE, default=0.0)
    cashflow_cum = Column(DOUBLE, default=0.0)
    rr = Column(DOUBLE, default=0.0)  # Return Rate
    rr_nc = Column(DOUBLE, default=0.0)  # Return Rate Compound
    rr_compound = Column(DOUBLE, default=0.0)  # Return Rate No Commission
    rr_compound_nc = Column(DOUBLE, default=0.0)  # Return Rate Compound No Commission
    calc_mode = Column(TINYINT)  # 计算模式：0 普通模式，1 保证金模式
    logger = logging.getLogger(f'<Table:{__tablename__}>')

    def __init__(self, stg_run_id=None, trade_agent_key=None,
                 trade_dt=None, trade_date=None, trade_time=None, trade_millisec=None, cash_available_last_day=0.0,
                 cash_available=0.0, position_value=0.0, curr_margin=0.0, close_profit=0.0, position_profit=0.0,
                 floating_pl_cum=0.0, cashflow_daily=0.0, cashflow_cum=0.0,
                 commission_tot=0.0, cash_init=0.0, calc_mode: (int, CalcMode) = CalcMode.Normal.value,
                 run_mode: (int, RunMode) = RunMode.Backtest.value):
        self.stg_run_id = stg_run_id
        self.trade_agent_status_detail_idx = None if stg_run_id is None else idx_generator(
            stg_run_id, TradeAgentStatusDetail)
        self.trade_agent_key = trade_agent_key.name if isinstance(trade_agent_key, ExchangeName) else trade_agent_key
        self.trade_dt = date_time_2_str(trade_date, trade_time) if trade_dt is None else trade_dt
        self.trade_date = trade_date
        self.trade_time = trade_time
        self.trade_millisec = trade_millisec
        self.cash_available_last_day = cash_available_last_day
        self.cash_available = cash_available
        self.position_value = position_value
        self.curr_margin = curr_margin
        self.close_profit = close_profit
        self.position_profit = position_profit
        self.floating_pl_cum = floating_pl_cum
        self.commission_tot = commission_tot
        self.cash_init = cash_init
        self.cash_and_margin = cash_available + curr_margin
        self.cashflow_daily = cashflow_daily
        self.cashflow_cum = cashflow_cum
        self.rr = 0  # Return Rate
        self.rr_compound = 0  # Return Rate Compound
        self.rr_nc = 0  # Return Rate No Commission
        self.rr_compound_nc = 0  # Return Rate Compound No Commission
        self.calc_mode = calc_mode.value if isinstance(calc_mode, CalcMode) else calc_mode
        self.run_mode = run_mode.value if isinstance(run_mode, RunMode) else run_mode
        self.pos_status_detail_dic = {}  # 用于记录当期状态对应的 pos_status_detail_dic
        self.last_status = None  # 用于记录上一个状态的实力

    def __repr__(self):
        return f"<TradeAgentStatusDetail(id='{self.trade_agent_status_detail_idx}', " \
               f"trade_agent_key={self.trade_agent_key}, trade_dt='{datetime_2_str(self.trade_dt)}', " \
               f"cash_available='{self.cash_available}', cash_available_last_day='{self.cash_available_last_day}', " \
               f"cashflow_daily='{self.cashflow_daily}', cashflow_cum='{self.cashflow_cum}', " \
               f"cash_and_margin='{self.cashflow_cum}', cash_and_margin='{self.cashflow_cum}', " \
               f"floating_pl_cum='{self.floating_pl_cum}')>"

    @staticmethod
    def create_t_1(stg_run_id, trade_agent_key, init_cash: int, timestamp_curr: (datetime, pd.Timestamp) = None,
                   md: dict = None, timestamp_key=None, date_key=None, time_key=None, milli_sec_key=None,
                   calc_mode: (int, CalcMode) = CalcMode.Normal.value,
                   run_mode: (int, RunMode) = RunMode.Backtest.value):
        """
        根据 md 及 初始化资金 创建对象，默认日期为当前md数据-1天
        :param stg_run_id:
        :param trade_agent_key:
        :param init_cash: 
        :param timestamp_curr: timestamp_curr 或 md + key 参数 两组参数二选其一
        :param md: timestamp_curr 或 md + key 参数 两组参数二选其一
        :param timestamp_key: timestamp_curr 或 md + key 参数 两组参数二选其一
        :param date_key: timestamp_curr 或 md + key 参数 两组参数二选其一
        :param time_key: timestamp_curr 或 md + key 参数 两组参数二选其一
        :param milli_sec_key: timestamp_curr 或 md + key 参数 两组参数二选其一
        :param calc_mode:
        :param run_mode:
        :return:
        """
        warnings.warn('该函数为范例函数，可以根据实际情况改写', UserWarning)
        if timestamp_curr is not None:
            timestamp_curr -= pd.Timedelta(days=1)
        elif timestamp_key is not None:
            timestamp_curr = md[timestamp_key] - pd.Timedelta(days=1)

        if timestamp_curr is None:
            trade_date = str_2_date(md[date_key]) - timedelta(days=1)
            trade_time = pd_timedelta_2_timedelta(md[time_key])
            trade_dt = datetime_2_str(date_time_2_str(trade_date, trade_time))
            trade_millisec = int(md.setdefault(milli_sec_key, 0)) if milli_sec_key is not None else 0
        else:
            trade_date = timestamp_curr.date()
            trade_time = timestamp_curr.time()
            trade_dt = timestamp_curr
            trade_millisec = int(md.setdefault(milli_sec_key, 0)) if milli_sec_key is not None else 0

        # calc_mode = calc_mode.value if isinstance(calc_mode, CalcMode) else calc_mode
        acc_status_detail = TradeAgentStatusDetail(stg_run_id=stg_run_id,
                                                   trade_agent_key=trade_agent_key,
                                                   trade_dt=trade_dt,
                                                   trade_date=trade_date,
                                                   trade_time=trade_time,
                                                   trade_millisec=trade_millisec,
                                                   cash_available_last_day=init_cash,
                                                   cash_available=init_cash,
                                                   position_value=0,
                                                   curr_margin=0,
                                                   close_profit=0,
                                                   position_profit=0,
                                                   cash_init=init_cash,
                                                   calc_mode=calc_mode,
                                                   run_mode=run_mode,
                                                   )
        if config.ORM_UPDATE_OR_INSERT_PER_ACTION:
            # 更新最新持仓纪录
            with with_db_session(engine_ibats, expire_on_commit=False) as session:
                session.add(acc_status_detail)
                session.commit()
        return acc_status_detail

    def create_by_self(self, is_new_day=False):
        """
        创建新的对象，默认前一日持仓信息的最新价，等于下一交易日的结算价（即 AvePrice）
        :return: 
        """
        cash_available_last_day = self.cash_available if is_new_day else self.cash_available_last_day
        detail = TradeAgentStatusDetail(stg_run_id=self.stg_run_id,
                                        trade_agent_key=self.trade_agent_key,
                                        trade_date=self.trade_date,
                                        trade_time=self.trade_time,
                                        trade_millisec=self.trade_millisec,
                                        cash_available_last_day=cash_available_last_day,
                                        cash_available=self.cash_available,
                                        curr_margin=self.curr_margin,
                                        close_profit=self.close_profit,
                                        position_profit=self.position_profit,
                                        floating_pl_cum=self.floating_pl_cum,
                                        cashflow_daily=self.cashflow_daily,
                                        cashflow_cum=self.cashflow_cum,
                                        commission_tot=self.commission_tot,
                                        cash_init=self.cash_init,
                                        calc_mode=self.calc_mode,
                                        run_mode=self.run_mode,
                                        )
        detail.last_status = self
        return detail

    def update_by_pos_status_detail(
            self, pos_status_detail_dic, timestamp_curr: (datetime, pd.Timestamp) = None, md: dict = None,
            timestamp_key=None, date_key=None, time_key=None, milli_sec_key=None):
        """
        根据 持仓列表更新账户信息
        :param pos_status_detail_dic:
        :param timestamp_curr: timestamp_curr 或 md + key 参数 两组参数二选其一
        :param md: timestamp_curr 或 md + key 参数 两组参数二选其一
        :param timestamp_key: timestamp_curr 或 md + key 参数 两组参数二选其一
        :param date_key: timestamp_curr 或 md + key 参数 两组参数二选其一
        :param time_key: timestamp_curr 或 md + key 参数 两组参数二选其一
        :param milli_sec_key: timestamp_curr 或 md + key 参数 两组参数二选其一
        :return:
        """
        if self.run_mode == RunMode.Backtest.value:
            detail = self._update_by_pos_status_detail(
                pos_status_detail_dic, timestamp_curr, md, timestamp_key, date_key, time_key, milli_sec_key)
        elif self.run_mode == RunMode.Backtest_FixPercent.value:
            detail = self._update_by_pos_status_detail_fix_percent(
                pos_status_detail_dic, timestamp_curr, md, timestamp_key, date_key, time_key, milli_sec_key)

        detail.pos_status_detail_dic = pos_status_detail_dic.copy()
        return detail

    def _update_by_pos_status_detail(self, pos_status_detail_dic, timestamp_curr: (datetime, pd.Timestamp) = None,
                                     md: dict = None,
                                     timestamp_key=None, date_key=None, time_key=None, milli_sec_key=None):
        """
        根据 持仓列表更新账户信息
        :param pos_status_detail_dic:
        :param timestamp_curr: timestamp_curr 或 md + key 参数 两组参数二选其一
        :param md: timestamp_curr 或 md + key 参数 两组参数二选其一
        :param timestamp_key: timestamp_curr 或 md + key 参数 两组参数二选其一
        :param date_key: timestamp_curr 或 md + key 参数 两组参数二选其一
        :param time_key: timestamp_curr 或 md + key 参数 两组参数二选其一
        :param milli_sec_key: timestamp_curr 或 md + key 参数 两组参数二选其一
        :return:
        """
        # warnings.warn('该函数为范例函数，可以根据实际情况改写', UserWarning)
        # 更新日期、时间
        if timestamp_curr is None:
            timestamp_curr = md[timestamp_key]

        if timestamp_curr is None:
            trade_date = str_2_date(md[date_key]) - timedelta(days=1)
            trade_time = pd_timedelta_2_timedelta(md[time_key])
            trade_dt = datetime_2_str(date_time_2_str(trade_date, trade_time))
            trade_millisec = int(md.setdefault(milli_sec_key, 0)) if milli_sec_key is not None else 0
        else:
            trade_date = timestamp_curr.date()
            trade_time = timestamp_curr.time()
            trade_dt = timestamp_curr
            trade_millisec = int(md.setdefault(milli_sec_key, 0)) if milli_sec_key is not None else 0

        is_new_day = self.trade_date != trade_date
        detail = self.create_by_self(is_new_day=is_new_day)
        detail.trade_dt = trade_dt
        detail.trade_date = trade_date
        detail.trade_time = trade_time
        detail.trade_millisec = trade_millisec

        # 更新非日期数据
        curr_margin = 0
        position_value = 0
        close_profit = 0
        position_profit = 0
        floating_pl_cum = 0
        cashflow_daily, cashflow_cum = 0, 0
        commission_tot = 0
        for instrument_id, pos_status_detail in pos_status_detail_dic.items():
            position_value += pos_status_detail.position_value
            curr_margin += pos_status_detail.margin
            position_profit += pos_status_detail.floating_pl
            close_profit += (pos_status_detail.floating_pl_cum - pos_status_detail.floating_pl)
            floating_pl_cum += pos_status_detail.floating_pl_cum
            cashflow_daily += pos_status_detail.cashflow_daily
            cashflow_cum += pos_status_detail.cashflow_cum
            commission_tot += pos_status_detail.commission_tot

        cash_available = self.cash_init + cashflow_cum

        # 记录当前可用现金
        detail.cash_available = cash_available
        cash_available_last_day = detail.cash_available_last_day
        calc_gap = cash_available - (cash_available_last_day + cashflow_daily)
        if calc_gap < -0.01 or 0.01 < calc_gap:
            self.logger.warning(
                "%s cash_init + cashflow_cum = %10.2f 比 cash_available_last_day + cashflow_daily = %10.2f 多 %10.2f。"
                "cash_init=%10.2f cashflow_cum=%10.2f cash_available_last_day=%10.2f cashflow_daily=%10.2f",
                detail.trade_dt,
                cash_available, cash_available_last_day + cashflow_daily,
                                cash_available - (cash_available_last_day + cashflow_daily),
                self.cash_init, cashflow_cum, cash_available_last_day, cashflow_daily)
        else:
            pass

        detail.position_value = position_value
        detail.curr_margin = curr_margin
        detail.close_profit = close_profit
        detail.position_profit = position_profit
        detail.floating_pl_cum = floating_pl_cum
        detail.cashflow_daily = cashflow_daily
        detail.cashflow_cum = cashflow_cum
        detail.commission_tot = commission_tot
        detail.cash_and_margin = detail.cash_available + curr_margin
        detail.rr = detail.floating_pl_cum / detail.cash_init
        detail.rr_nc = (detail.floating_pl_cum + commission_tot) / detail.cash_init
        # 当期盈利及现金的计算均是按照单利计算的，计算rr的时候需要将单利转化为复利
        # 计算方法为：
        # 当期状态复利rr = ("当期状态的单利 rr" - "上一状态的单利 rr" + 1) * ("上一状态的复利 rr" + 1) - 1
        # 进一步合并公式 =  （“当期 cash_and_margin” - “上一状态 cash_and_margin” + cash_init）/ cash_init * "上一状态的复利 rr"
        detail.rr_compound = (detail.rr - self.rr + 1) * (self.rr + 1) - 1
        detail.rr_compound_nc = (detail.rr_nc - self.rr_nc + 1) * (self.rr_nc + 1) - 1

        if config.ORM_UPDATE_OR_INSERT_PER_ACTION:
            # 更新最新持仓纪录
            with with_db_session(engine_ibats, expire_on_commit=False) as session:
                session.add(detail)
                session.commit()
        return detail

    def _update_by_pos_status_detail_fix_percent(self, pos_status_detail_dic,
                                                 timestamp_curr: (datetime, pd.Timestamp) = None, md: dict = None,
                                                 timestamp_key=None, date_key=None, time_key=None, milli_sec_key=None):
        """
        根据 持仓列表更新账户信息
        :param pos_status_detail_dic:
        :param timestamp_curr: timestamp_curr 或 md + key 参数 两组参数二选其一
        :param md: timestamp_curr 或 md + key 参数 两组参数二选其一
        :param timestamp_key: timestamp_curr 或 md + key 参数 两组参数二选其一
        :param date_key: timestamp_curr 或 md + key 参数 两组参数二选其一
        :param time_key: timestamp_curr 或 md + key 参数 两组参数二选其一
        :param milli_sec_key: timestamp_curr 或 md + key 参数 两组参数二选其一
        :return:
        """
        # warnings.warn('该函数为范例函数，可以根据实际情况改写', UserWarning)
        # 更新日期、时间
        if timestamp_curr is None:
            timestamp_curr = md[timestamp_key]

        if timestamp_curr is None:
            trade_date = str_2_date(md[date_key]) - timedelta(days=1)
            trade_time = pd_timedelta_2_timedelta(md[time_key])
            trade_dt = datetime_2_str(date_time_2_str(trade_date, trade_time))
            trade_millisec = int(md.setdefault(milli_sec_key, 0)) if milli_sec_key is not None else 0
        else:
            trade_date = timestamp_curr.date()
            trade_time = timestamp_curr.time()
            trade_dt = timestamp_curr
            trade_millisec = int(md.setdefault(milli_sec_key, 0)) if milli_sec_key is not None else 0

        is_new_day = self.trade_date != trade_date
        detail = self.create_by_self(is_new_day=is_new_day)
        detail.trade_dt = trade_dt
        detail.trade_date = trade_date
        detail.trade_time = trade_time
        detail.trade_millisec = trade_millisec

        # 上一状态时的 pos_status_detail_dic
        # pos_status_detail_dic_last = self.last_status.pos_status_detail_dic if self.last_status is not None else {}
        pos_status_detail_dic_last = self.pos_status_detail_dic
        # 更新非日期数据
        curr_margin = 0
        position_value = 0
        close_profit = 0
        position_profit = 0
        floating_pl_cum = 0
        cashflow_daily, cashflow_cum = 0, 0
        commission_tot = 0
        position_rate_tot = 0
        for instrument_id, pos_status_detail in pos_status_detail_dic.items():
            position_rate = pos_status_detail.position
            curr_margin += position_rate  # curr_margin 记录的是持仓比例例如 1 满仓; 0 空仓;
            position_rate_tot += position_rate  # 累计持仓比例 1 满仓; 0 空仓;
            # 固定比例仓位下，固定的是“margin 固定保证金比例”，而不是 “position_value 持仓手数”，因此，其他数据的计算需要进行一次转换
            # margin_2_pos_rate 是用来记录 从在固定 margin 的情况下，其他数据的转换比例
            if instrument_id in pos_status_detail_dic_last:
                pos_status_detail_last = pos_status_detail_dic_last[instrument_id]
                if pos_status_detail.margin > 0:
                    margin_2_pos_rate = position_rate / pos_status_detail.margin
                elif pos_status_detail.margin == 0 and pos_status_detail_last.margin > 0:
                    margin_2_pos_rate = position_rate / pos_status_detail_last.margin
                else:
                    margin_2_pos_rate = 0

                # 计算 close_profit, floating_pl_cum, cashflow_cum, commission_tot 需要与上一状态插值进行计算
                close_profit += ((pos_status_detail.floating_pl_cum - pos_status_detail.floating_pl) -
                                 (pos_status_detail_last.floating_pl_cum - pos_status_detail_last.floating_pl)
                                 ) * margin_2_pos_rate
                floating_pl_cum += (pos_status_detail.floating_pl_cum - pos_status_detail_last.floating_pl_cum
                                    ) * margin_2_pos_rate
                cashflow_cum += (pos_status_detail.cashflow_cum - pos_status_detail_last.cashflow_cum
                                 ) * margin_2_pos_rate
                commission_tot += (pos_status_detail.commission_tot - pos_status_detail_last.commission_tot
                                   ) * margin_2_pos_rate

            else:
                if pos_status_detail.margin > 0:
                    margin_2_pos_rate = position_rate / pos_status_detail.margin
                else:
                    margin_2_pos_rate = 0

                # 计算 close_profit, floating_pl_cum, cashflow_cum, commission_tot 需要与上一状态插值进行计算
                close_profit += (pos_status_detail.floating_pl_cum - pos_status_detail.floating_pl) * margin_2_pos_rate
                floating_pl_cum += pos_status_detail.floating_pl_cum * margin_2_pos_rate
                cashflow_cum += pos_status_detail.cashflow_cum * margin_2_pos_rate
                commission_tot += pos_status_detail.commission_tot * margin_2_pos_rate

            # 计算position_value, cashflow_daily, position_profit
            position_value += pos_status_detail.position_value * margin_2_pos_rate
            cashflow_daily += pos_status_detail.cashflow_daily * margin_2_pos_rate
            position_profit += pos_status_detail.floating_pl * margin_2_pos_rate

        # 计算 close_profit, floating_pl_cum, cashflow_cum, commission_tot 需要与上一状态插值进行计算
        close_profit = self.close_profit + close_profit
        floating_pl_cum = self.floating_pl_cum + floating_pl_cum
        cashflow_cum = self.cashflow_cum + cashflow_cum
        commission_tot = self.commission_tot + commission_tot

        # 这里不考虑手续费造成的实际可用现金可能为负数的清空，仅简单计算 cash + margin = 1
        # cash_available = 1 - curr_margin

        # 记录当前可用现金，在补丁比例模式下，可用现金等于初始现金 - 累计持仓比例
        detail.cash_available = cash_available = self.cash_init - position_rate_tot
        # 以下检查不适用。因为固定比例持仓分析，cash_available 相当于闲置现金比例，不会随盈利现金增长而等比例变化
        # cash_available_last_day = detail.cash_available_last_day
        # calc_gap = cash_available - (cash_available_last_day + cashflow_daily)
        # if calc_gap < -0.01 or 0.01 < calc_gap:
        #     self.logger.warning(
        #         "%s cash_init + cashflow_cum = %10.5f 比 cash_available_last_day + cashflow_daily = %10.5f 多 %10.5f。"
        #         "cash_init=%10.5f cashflow_cum=%10.5f cash_available_last_day=%10.5f cashflow_daily=%10.5f",
        #         detail.trade_dt,
        #         cash_available, cash_available_last_day + cashflow_daily,
        #                         cash_available - (cash_available_last_day + cashflow_daily),
        #         self.cash_init, cashflow_cum, cash_available_last_day, cashflow_daily)

        detail.position_value = position_value
        detail.curr_margin = curr_margin
        detail.close_profit = close_profit
        detail.position_profit = position_profit
        detail.floating_pl_cum = floating_pl_cum
        detail.cashflow_daily = cashflow_daily
        detail.cashflow_cum = cashflow_cum
        detail.commission_tot = commission_tot
        detail.cash_and_margin = self.cash_init + cashflow_cum + curr_margin
        detail.rr = detail.floating_pl_cum / detail.cash_init
        detail.rr_nc = (detail.floating_pl_cum + commission_tot) / detail.cash_init
        # 当期盈利及现金的计算均是按照单利计算的，计算rr的时候需要将单利转化为复利
        # 计算方法为：
        # 当期状态复利rr = ("当期状态的单利 rr" - "上一状态的单利 rr" + 1) * ("上一状态的复利 rr" + 1) - 1
        # 进一步合并公式 =  （“当期 cash_and_margin” - “上一状态 cash_and_margin” + cash_init）/ cash_init * "上一状态的复利 rr"
        detail.rr_compound = (detail.rr - self.rr + 1) * (self.rr + 1) - 1
        detail.rr_compound_nc = (detail.rr_nc - self.rr_nc + 1) * (self.rr_nc + 1) - 1

        if config.ORM_UPDATE_OR_INSERT_PER_ACTION:
            # 更新最新持仓纪录
            with with_db_session(engine_ibats, expire_on_commit=False) as session:
                session.add(detail)
                session.commit()
        return detail


class StgRunStatusDetail(BaseModel):
    """记录每一单位时刻 所有trade_agent_detail 状态汇总"""
    __tablename__ = 'stg_run_status_detail'
    stg_run_id = Column(Integer, primary_key=True)  # 对应回测了策略 StgRunID 此数据与 AccSumID 对应数据相同
    stg_run_status_detail_idx = Column(Integer, primary_key=True)
    trade_dt = Column(DateTime)
    trade_date = Column(Date)  # 对应行情数据中 ActionDate
    trade_time = Column(Time)  # 对应行情数据中 ActionTime
    trade_millisec = Column(Integer)  # 对应行情数据中 ActionMillisec
    cash_available_last_day = Column(DOUBLE, default=0.0)  # 可用资金, double
    cash_available = Column(DOUBLE, default=0.0)  # 可用资金, double
    position_value = Column(DOUBLE, default=0.0)
    curr_margin = Column(DOUBLE, default=0.0)  # 当前保证金总额, double
    close_profit = Column(DOUBLE, default=0.0)
    position_profit = Column(DOUBLE, default=0.0)
    floating_pl_cum = Column(DOUBLE, default=0.0)
    commission_tot = Column(DOUBLE, default=0.0)
    cash_init = Column(DOUBLE, default=0.0)
    cash_and_margin = Column(DOUBLE, default=0.0)
    cashflow_daily = Column(DOUBLE, default=0.0)
    cashflow_cum = Column(DOUBLE, default=0.0)
    rr = Column(DOUBLE, default=0.0)  # Return Rate
    rr_nc = Column(DOUBLE, default=0.0)  # Return Rate Compound
    rr_compound = Column(DOUBLE, default=0.0)  # Return Rate No Commission
    rr_compound_nc = Column(DOUBLE, default=0.0)  # Return Rate Compound No Commission

    def __init__(self, stg_run_id=None,
                 trade_dt=None, trade_date=None, trade_time=None, trade_millisec=None, cash_available_last_day=0.0,
                 cash_available=None, curr_margin=None, close_profit=None, position_profit=None, floating_pl_cum=None,
                 commission_tot=None, cash_init=None, cashflow_daily=0.0, cashflow_cum=0.0,
                 rr=0, rr_nc=0.0, rr_compound=0.0, rr_compound_nc=0.0):
        self.stg_run_id = stg_run_id
        self.stg_run_status_detail_idx = None if stg_run_id is None else idx_generator(
            stg_run_id, StgRunStatusDetail)
        self.trade_dt = date_time_2_str(trade_date, trade_time) if trade_dt is None else trade_dt
        self.trade_date = trade_date
        self.trade_time = trade_time
        self.trade_millisec = trade_millisec
        self.cash_available_last_day = cash_available_last_day
        self.cash_available = cash_available
        self.curr_margin = curr_margin
        self.close_profit = close_profit
        self.position_profit = position_profit
        self.floating_pl_cum = floating_pl_cum
        self.commission_tot = commission_tot
        self.cash_init = cash_init
        self.cash_and_margin = cash_available + curr_margin
        self.cashflow_daily = cashflow_daily
        self.cashflow_cum = cashflow_cum
        self.rr = rr
        self.rr_nc = rr_nc
        self.rr_compound = rr_compound
        self.rr_compound_nc = rr_compound_nc

    @staticmethod
    def create_t_1(
            stg_run_id, trade_agent_status_detail_list):
        """通过 trade_agent_status_detail_list 更新当前 stg_run_detail 状态"""
        data_len = len(trade_agent_status_detail_list)
        if data_len == 0:
            return None

        trade_dt, trade_date, trade_time, trade_millisec = None, None, None, None
        cash_available_last_day, cash_available, curr_margin, close_profit, position_profit = 0, 0, 0, 0, 0
        floating_pl_cum, commission_tot, cash_init, cash_and_margin, cashflow_daily, cashflow_cum, rr = 0, 0, 0, 0, 0, 0, 0
        for detail in trade_agent_status_detail_list:
            if trade_dt is None or trade_dt < detail.trade_dt:
                trade_dt = detail.trade_dt
                trade_date = detail.trade_date
                trade_time = detail.trade_time
                trade_millisec = detail.trade_millisec

            cash_available_last_day += 0 if detail.cash_available_last_day is None else detail.cash_available_last_day
            cash_available += 0 if detail.cash_available is None else detail.cash_available
            curr_margin += 0 if detail.curr_margin is None else detail.curr_margin
            close_profit += 0 if detail.close_profit is None else detail.close_profit
            position_profit += 0 if detail.position_profit is None else detail.position_profit
            floating_pl_cum += 0 if detail.floating_pl_cum is None else detail.floating_pl_cum
            commission_tot += 0 if detail.commission_tot is None else detail.commission_tot
            cash_init += 0 if detail.cash_init is None else detail.cash_init
            cash_and_margin += 0 if detail.cash_and_margin is None else detail.cash_and_margin
            cashflow_daily += 0 if detail.cashflow_daily is None else detail.cashflow_daily
            cashflow_cum += 0 if detail.cashflow_cum is None else detail.cashflow_cum

        detail = StgRunStatusDetail(
            stg_run_id=stg_run_id,
            trade_dt=trade_dt,
            trade_date=trade_date,
            trade_time=trade_time,
            trade_millisec=trade_millisec,
            cash_available_last_day=cash_available_last_day,
            cash_available=cash_available,
            curr_margin=curr_margin,
            close_profit=close_profit,
            position_profit=position_profit,
            floating_pl_cum=floating_pl_cum,
            commission_tot=commission_tot,
            cash_init=cash_init,
            cashflow_daily=cashflow_daily,
            cashflow_cum=cashflow_cum,
        )

        return detail

    def update_by_trade_agent_status_detail_list(self, trade_agent_status_detail_list):
        data_len = len(trade_agent_status_detail_list)
        if data_len == 0:
            return None

        trade_dt, trade_date, trade_time, trade_millisec = None, None, None, None
        cash_available_last_day, cash_available, curr_margin, close_profit, position_profit = 0, 0, 0, 0, 0
        floating_pl_cum, commission_tot, cash_init, cash_and_margin, cashflow_daily, cashflow_cum, rr = 0, 0, 0, 0, 0, 0, 0
        for detail in trade_agent_status_detail_list:
            if trade_dt is None or trade_dt < detail.trade_dt:
                trade_dt = detail.trade_dt
                trade_date = detail.trade_date
                trade_time = detail.trade_time
                trade_millisec = detail.trade_millisec

            cash_available_last_day += 0 if detail.cash_available_last_day is None else detail.cash_available_last_day
            cash_available += 0 if detail.cash_available is None else detail.cash_available
            curr_margin += 0 if detail.curr_margin is None else detail.curr_margin
            close_profit += 0 if detail.close_profit is None else detail.close_profit
            position_profit += 0 if detail.position_profit is None else detail.position_profit
            floating_pl_cum += 0 if detail.floating_pl_cum is None else detail.floating_pl_cum
            commission_tot += 0 if detail.commission_tot is None else detail.commission_tot
            cash_init += 0 if detail.cash_init is None else detail.cash_init
            cash_and_margin += 0 if detail.cash_and_margin is None else detail.cash_and_margin
            cashflow_daily += 0 if detail.cashflow_daily is None else detail.cashflow_daily
            cashflow_cum += 0 if detail.cashflow_cum is None else detail.cashflow_cum

        rr = floating_pl_cum / cash_init
        rr_nc = (floating_pl_cum + commission_tot) / cash_init
        # 当期盈利及现金的计算均是按照单利计算的，计算rr的时候需要将单利转化为复利
        # 计算方法为：
        # 当期状态复利rr = ("当期状态的单利 rr" - "上一状态的单利 rr" + 1) * ("上一状态的复利 rr" + 1) - 1
        # 进一步合并公式 =  （“当期 cash_and_margin” - “上一状态 cash_and_margin” + cash_init）/ cash_init * "上一状态的复利 rr"
        rr_compound = (rr - self.rr + 1) * (self.rr + 1) - 1
        rr_compound_nc = (rr_nc - self.rr_nc + 1) * (self.rr_nc + 1) - 1

        detail = StgRunStatusDetail(
            stg_run_id=self.stg_run_id,
            trade_dt=trade_dt,
            trade_date=trade_date,
            trade_time=trade_time,
            trade_millisec=trade_millisec,
            cash_available_last_day=cash_available_last_day,
            cash_available=cash_available,
            curr_margin=curr_margin,
            close_profit=close_profit,
            position_profit=position_profit,
            floating_pl_cum=floating_pl_cum,
            commission_tot=commission_tot,
            cash_init=cash_init,
            cashflow_daily=cashflow_daily,
            cashflow_cum=cashflow_cum,
            rr=rr,
            rr_nc=rr_nc,
            rr_compound=rr_compound,
            rr_compound_nc=rr_compound_nc,
        )
        return detail


def init():
    from ibats_utils.db import alter_table_2_myisam
    global engine_ibats
    engine_ibats = engines.engine_ibats
    BaseModel.metadata.create_all(engine_ibats)
    alter_table_2_myisam(engine_ibats)
    print("所有表结构建立完成")


if __name__ == "__main__":
    init()
    # 创建user表，继承metadata类
    # Engine使用Schama Type创建一个特定的结构对象
    # stg_info_table = Table("stg_info", metadata, autoload=True)



#! /usr/bin/env python
# -*- coding:utf-8 -*-
"""
@author  : MG
@Time    : 2018/6/20 15:16
@File    : __init__.py.py
@contact : mmmaaaggg@163.com
@desc    : 
"""
from ibats_common.config import config
from ibats_utils.db import DynamicEngine
import logging
from datetime import datetime

logger = logging.getLogger(__name__)


class DyEngine(DynamicEngine):

    def __init__(self, db_url_dic):
        DynamicEngine.__init__(self, db_url_dic)
        self.update_datetime = datetime.now()

    def __getitem__(self, item):
        if config.update_db_datetime > self.update_datetime:
            self.reload_engines(refresh=True)

        return DynamicEngine.__getitem__(self, item)

    def __iter__(self):
        if config.update_db_datetime > self.update_datetime:
            self.reload_engines(refresh=True)

        return DynamicEngine.__iter__(self)

    @property
    def engine_ibats(self):
        return self.__getitem__(config.DB_SCHEMA_IBATS)


engines = DyEngine(config.DB_URL_DIC)



#! /usr/bin/env python
# -*- coding:utf-8 -*-
"""
@author  : MG
@Time    : 19-4-19 上午8:40
@File    : ai_stg.py
@contact : mmmaaaggg@163.com
@desc    : 简单的 RNN LSTM 构建策略模型，运行该模型需要首先安装 TensorFlow 包
pip3 install tensorflow
"""
import os
from ibats_utils.mess import get_last_idx, get_folder_path
import tensorflow as tf
import numpy as np
import random
from ibats_common.common import PeriodType, RunMode, BacktestTradeMode, ExchangeName, ContextKey, Direction, CalcMode
from ibats_common.strategy import StgBase
from ibats_common.strategy_handler import strategy_handler_factory
from ibats_local_trader.agent.td_agent import *
from ibats_local_trader.agent.md_agent import *

logger = logging.getLogger(__name__)


class LSTMRNN:
    def __init__(self, n_step, n_inputs, n_hidden_units, n_classes, lr, batch_size, normalization_model):
        """

        :param n_step: time steps
        :param n_inputs: MNIST data input (img shape 28*28)
        :param n_hidden_units: neurons in hidden layer
        :param n_classes: MNIST classes (0-9 digits)
        :param lr:
        :param training_iters:
        :param batch_size:
        :param normalization_model:
        """
        self.n_step = n_step
        self.n_inputs = n_inputs
        self.n_hidden_units = n_hidden_units
        self.n_classes = n_classes
        # hyperparameters
        self.lr = lr
        self.batch_size = batch_size
        self.normalization_model = normalization_model
        # attributes defined in other functions
        self.cost = None
        self.l_in_y = None
        self.cell_outputs = None
        self.cell_init_state = None
        self.cell_final_state = None

        with tf.name_scope('inputs'):
            # tf Graph input
            self.xs = tf.placeholder(tf.float32, [None, n_step, n_inputs])
            self.ys = tf.placeholder(tf.float32, [None, n_classes])
            self.is_training = tf.placeholder(tf.bool, [])

        # Define weights
        self.weights = {
            # 28*128
            'in': tf.Variable(tf.random_normal([n_inputs, n_hidden_units])),
            # (128, 10)
            'out': tf.Variable(tf.random_normal([n_hidden_units, n_classes]))
        }

        self.biases = {
            # (128, )
            'in': tf.Variable(tf.constant(0.1, shape=[n_hidden_units, ])),
            # (10, )
            'out': tf.Variable(tf.constant(0.1, shape=[n_classes, ]))
        }
        with tf.variable_scope('in_hidden'):
            self.add_input_layer()
        with tf.variable_scope('LSTM_cell'):
            self.add_cell()
        with tf.variable_scope('out_hidden'):
            self.add_output_layer()
        with tf.name_scope('cost'):
            self.compute_cost()
        with tf.name_scope('train'):
            self.train_op = tf.train.AdamOptimizer(self.lr).minimize(self.cost)
        with tf.name_scope('accuracy'):
            correct_pred = tf.equal(tf.argmax(self.pred, 1), tf.argmax(self.ys, 1))
            self.accuracy_op = tf.reduce_mean(tf.cast(correct_pred, tf.float32))

    def add_input_layer(self):
        # hidden layer for input to cell
        # X (128 batch, 28 steps, 28 inputs)
        # ==> X (128 * 28, 28 inputs)
        self.l_in_x = tf.reshape(self.xs, [-1, self.n_inputs])
        self.Ws_in = tf.Variable(tf.random_normal([self.n_inputs, self.n_hidden_units]))
        self.bs_in = tf.Variable(tf.constant(0.1, shape=[self.n_hidden_units, ]))
        # ==> X_in (128 batch * 28 steps, 128 hidden)
        with tf.name_scope('Wx_plus_b'):
            l_in_y = tf.matmul(self.l_in_x, self.Ws_in) + self.bs_in
            if self.normalization_model:
                l_in_y = tf.layers.batch_normalization(l_in_y, training=self.is_training)

        # ==> X_in (128 batch, 28 steps, 128 hidden)
        self.l_in_y = tf.reshape(l_in_y, [-1, self.n_step, self.n_hidden_units])

    def add_cell(self):
        # cell
        lstm_cell = tf.nn.rnn_cell.BasicLSTMCell(self.n_hidden_units, forget_bias=1.0, state_is_tuple=True)
        self.cell_outputs, self.cell_final_state = tf.nn.dynamic_rnn(
            lstm_cell, self.l_in_y, time_major=False, dtype=tf.float32)

    def add_output_layer(self):
        # hidden layer for output as the final results
        # method 1
        # results = tf.matmul(states[1], weights['out']) + biases['out']  # states[1] is m_state
        # method 2
        # unpack to list[(batch, outputs)...] * steps
        self.l_out_x = tf.unstack(tf.transpose(self.cell_outputs, [1, 0, 2]))  # states is the last outputs
        self.Ws_out = tf.Variable(tf.random_normal([self.n_hidden_units, self.n_classes]))
        self.bs_out = tf.Variable(tf.constant(0.1, shape=[self.n_classes, ]))
        self.pred = tf.matmul(self.l_out_x[-1], self.Ws_out) + self.bs_out

    def compute_cost(self):
        self.cost = tf.reduce_mean(tf.nn.softmax_cross_entropy_with_logits_v2(
            logits=self.pred, labels=self.ys))
        tf.summary.scalar('cost', self.cost)


class AIStg(StgBase):

    def __init__(self, unit=1, train=True):
        super().__init__()
        self.unit = unit
        self.input_size = 13
        self.batch_size = 50
        self.n_step = 20
        self.output_size = 2
        self.n_hidden_units = 10
        self.lr = 0.006
        self.normalization_model = True
        self._model = None
        # tf.Session()
        self._session = None
        self.train_validation_rate = 0.8
        self.is_load_model_if_exist = True
        folder_path = get_folder_path('my_net', create_if_not_found=False)
        file_path = os.path.join(folder_path, f"save_net_{self.normalization_model}.ckpt")
        self.model_file_path = file_path
        self.training_iters = 600

    @property
    def session(self):
        if self._session is None:
            if self.model is None:
                raise ValueError('model 需要先于 session 被创建')
            self._session = tf.Session()
            self._session.run(tf.global_variables_initializer())
        return self._session

    @property
    def model(self):
        if self._model is None:
            self._model = self.build_model()
        return self._model

    def get_factors(self, md_df: pd.DataFrame, tail_n=None):
        if tail_n is not None:
            md_df = md_df.tail(tail_n)
        df = md_df[~md_df['close'].isnull()][[
            'open', 'high', 'low', 'close', 'volume', 'oi', 'warehousewarrant', 'termstructure']]
        df['ma5'] = df['close'].rolling(window=5).mean()
        df['ma10'] = df['close'].rolling(window=10).mean()
        df['ma20'] = df['close'].rolling(window=20).mean()
        df['pct_change_vol'] = df['volume'].pct_change()
        df['pct_change'] = df['close'].pct_change()

        factors = df.fillna(0).to_numpy()
        if self.input_size is None or self.input_size != factors.shape[1]:
            self.input_size = factors.shape[1]
            self.n_hidden_units = self.input_size * 2
            logger.info("set input_size: %d", self.input_size)
            logger.info("set n_hidden_units: %d", self.n_hidden_units)

        # if self.normalization_model:
        #     factors = (factors - np.mean(factors, 0)) / np.std(factors, 0)

        return factors

    def get_factors_with_labels(self, md_df):
        factors = self.get_factors(md_df)
        price_arr = factors[:, 0]
        self.input_size = factors.shape[1]
        labels = self.calc_label_with_future_value(price_arr, -0.01, 0.01)
        idx_last_available_label = get_last_idx(labels, lambda x: x.sum() == 0)
        factors = factors[:idx_last_available_label + 1, :]
        labels = labels[:idx_last_available_label + 1, :]
        # if self.normalization_model:
        #     factors = (factors - np.mean(factors, 0)) / np.std(factors, 0)

        return factors, labels

    def calc_label_with_future_value(self, value_arr: np.ndarray, min_pct: float, max_pct: float, max_future=None):
        """
        根据时间序列数据 pct_arr 计算每一个时点目标标示 -1 0 1
        计算方式：
        当某一点未来波动首先 > 上届 min_pct，则标记为： [0, 1]
        当某一点未来波动首先 < 下届 max_pct，则标记为： [1, 0]
        :param value_arr:
        :param min_pct:
        :param max_pct:
        :param max_future:最大搜索长度
        :param output_size:最大搜索长度
        :return:
        """

        value_arr[np.isnan(value_arr)] = 0
        arr_len = value_arr.shape[0]
        target_arr = np.zeros((arr_len, self.output_size))
        for i in range(arr_len):
            base = value_arr[i]
            for j in range(i + 1, arr_len):
                result = value_arr[j] / base - 1
                if result < min_pct:
                    target_arr[i, 0] = 1
                    break
                elif result > max_pct:
                    target_arr[i, 1] = 1
                    break
        return target_arr

    def get_batch_xs(self, factors: np.ndarray, index=None):
        """
        取 batch_xs
        :param factors:
        :param index: 样本起始坐标，如果为None，则默认取尾部一组样本
        :return:
        """
        if index is None:
            index = factors.shape[0] - 1

        batch_xs = np.zeros((1, self.n_step, self.input_size))
        batch_xs[0, :, :] = factors[(index - self.n_step + 1):(index + 1), :]

        return batch_xs

    def get_batch_by_random(self, factors: np.ndarray, labels: np.ndarray):
        """
        够在一系列输入输出数据集
        xs： 两条同频率，不同位移的sin曲线
        ys_value： 目标是一条cos曲线
        ys: ys_value 未来涨跌标识
        i_s：X 序列
        """
        xs = np.zeros((self.batch_size, self.n_step, self.input_size))
        ys = np.zeros((self.batch_size, self.output_size))
        # available_batch_size, num = 0, 0
        samples_index = random.sample(range(self.n_step - 1, factors.shape[0] - 1), self.batch_size)
        examples_index_list = []
        for available_batch_size, index in enumerate(samples_index):
            tmp = factors[(index - self.n_step + 1):(index + 1), :]
            if tmp.shape[0] < self.n_step:
                break
            xs[available_batch_size, :, :] = tmp
            ys[available_batch_size, :] = labels[index, :]
            examples_index_list.append(index)
            if available_batch_size + 1 >= self.batch_size:
                available_batch_size += 1
                break

        # returned xs, ys_value and shape (batch, step, input)
        return xs, ys, examples_index_list

    def build_model(self):
        # hyperparameters
        # lr = LR
        # batch_size = BATCH_SIZE
        #
        # n_inputs = INPUT_SIZE  # MNIST data input (img shape 28*28)
        # n_step = TIME_STEPS  # time steps
        # n_hidden_units = CELL_SIZE  # neurons in hidden layer
        # n_classes = OUTPUT_SIZE  # MNIST classes (0-9 digits)
        model = LSTMRNN(
            self.n_step, self.input_size, self.n_hidden_units, self.output_size, self.lr, self.batch_size,
            self.normalization_model)
        return model

    def train(self, md_df):
        factors, labels = self.get_factors_with_labels(md_df)
        factors_train, factors_validation, labels_train, labels_validation = self.separate_train_validation(
            factors, labels)

        sess = self.session
        merged = tf.summary.merge_all()
        writer = tf.summary.FileWriter('logs', sess.graph)
        # relocate to the local dir and run this line to view it on Chrome(http://0.0.0.0:6006/):
        # $ tensorboard --logdir='logs'

        step = 0
        while step < self.training_iters:  # * self.model.batch_size
            # batch_xs, batch_ys = mnist.train.next_batch(model.batch_size)
            # batch_xs.shape, batch_ys.shape
            batch_xs, batch_ys, examples_index_list = self.get_batch_by_random(factors_train, labels_train)
            # print("available_batch_size", available_batch_size)
            if len(examples_index_list) < self.model.batch_size:
                break
            # batch_xs = batch_xs.reshape([model.batch_size, model.n_step, model.n_inputs])
            # feed_dict = {model.xs: batch_xs, model.ys: batch_ys}
            # sess.run(model.train_op, feed_dict=feed_dict)
            feed_dict = {
                self.model.xs: batch_xs,
                self.model.ys: batch_ys,
                self.model.is_training: True,
            }

            # sess.run(model.train_op, feed_dict=feed_dict)
            _, cost, state, pred = sess.run(
                [self.model.train_op, self.model.cost, self.model.cell_final_state, self.model.pred]
                , feed_dict=feed_dict
            )

            if step % 20 == 0:
                train_accuracy = np.mean(np.argmax(pred, 1) == np.argmax(batch_ys, 1))
                batch_xs, batch_ys, _ = self.get_batch_by_random(factors_validation, labels_validation)
                feed_dict = {
                    self.model.xs: batch_xs,
                    self.model.ys: batch_ys,
                    self.model.is_training: False,
                    # TODO: model.is_training should be False
                }
                test_accuracy = sess.run(self.model.accuracy_op, feed_dict=feed_dict)
                logger.info('%d/%d) train: %s test: %s', step, self.training_iters, train_accuracy, test_accuracy)
                result = sess.run(merged, feed_dict)
                writer.add_summary(result, step)

            step += 1

        return self.model

    def save_model(self):
        """
        将模型导出到文件
        :return:
        """
        saver = tf.train.Saver(tf.trainable_variables(), max_to_keep=5)
        save_path = saver.save(self.session, self.model_file_path)
        logger.info("模型保存到: %s", save_path)
        return save_path

    def load_model_if_exist(self):
        """
        将模型导出到文件
        :return:
        """
        if self.is_load_model_if_exist and self.model_file_exists():
            # 检查文件是否存在
            model = self.model      # 这句话是必须的，需要实现建立模型才可以加载
            sess = self.session
            saver = tf.train.Saver(tf.trainable_variables())
            save_path = saver.restore(sess, self.model_file_path)
            logger.info("load from path: %s", save_path)
            return True

        return False

    def predict_test(self, md_df):
        logger.info('开始预测')
        sess = self.session
        # saver = tf.train.Saver()
        # saver.restore(sess, f"my_net/save_net_{self.model.normalization_model}.ckpt")
        factors, labels = self.get_factors_with_labels(md_df)
        logger.info("批量预测")
        batch_xs, batch_ys, examples_index_list = self.get_batch_by_random(factors, labels)
        feed_dict = {
            self.model.xs: batch_xs,
            self.model.ys: batch_ys,
            self.model.is_training: False,
            # TODO: model.is_training should be False
        }
        pred = sess.run(tf.argmax(self.model.pred, 1), feed_dict)
        logger.info("accuracy: %.2f%%" % (sum(pred == np.argmax(batch_ys, axis=1)) / len(pred) * 100))
        logger.info("pred: \n%s\n%s", pred, np.argmax(batch_ys, axis=1))

        logger.info("独立样本预测")
        # batch_xs, batch_ys, available_batch_size = self.get_batch_by_random(factors, labels)
        pred_all = []
        for n in range(len(examples_index_list)):
            feed_dict = {
                self.model.xs: batch_xs[n:n + 1, :, :],
                self.model.ys: batch_ys[n:n + 1, :],
                self.model.is_training: False,
                # TODO: model.is_training should be False
            }
            pred = sess.run(tf.argmax(self.model.pred, 1), feed_dict)
            pred_all.extend(pred)

        logger.info("accuracy: %.2f%%" % (sum(pred_all == np.argmax(batch_ys, axis=1)) / len(pred_all) * 100))
        logger.info("pred: \n%s\n%s", np.array(pred_all), np.argmax(batch_ys, axis=1))

        logger.info("独立样本预测(predict_latest)")
        # batch_xs, batch_ys, available_batch_size = self.get_batch_by_random(factors, labels)
        pred_all, label_all, label_target = [], [], np.argmax(labels, axis=1)
        for num, index in enumerate(examples_index_list):
            pred_mark = self.predict_latest(md_df.iloc[:(index + 1), :])
            pred_all.append(pred_mark)
            label_all.append(label_target[index])
            if pred_mark != pred_all[num] or label_target[index] != np.argmax(batch_ys, axis=1)[num]:
                logger.debug("(%d %d) (%d %d)",
                             pred_mark, label_target[index], pred_all[num], np.argmax(batch_ys, axis=1)[num])

        pred_all, label_all = np.array(pred_all), np.array(label_all)
        logger.info("accuracy: %.2f%%" % (sum(pred_all == label_all) / len(pred_all) * 100))
        logger.info("pred: \n%s\n%s", pred_all, label_all)

    def predict_latest(self, md_df):
        factors = self.get_factors(md_df, tail_n=self.n_step)
        batch_xs = self.get_batch_xs(factors)
        feed_dict = {
            self.model.xs: batch_xs,
            self.model.is_training: False,
            # TODO: model.is_training should be False
        }
        pred_mark = self.session.run(tf.argmax(self.model.pred, 1), feed_dict)[0]
        # is_buy, is_sell = pred_mark == 1, pred_mark == 0
        # return is_buy, is_sell
        return pred_mark

    def on_prepare_min1(self, md_df, context):
        if md_df is None:
            return

        # 加载模型
        is_load = self.load_model_if_exist()

        if not is_load:
            # 训练模型
            self.train(md_df)
            self.predict_test(md_df)
            self.save_model()

    def on_min1(self, md_df, context):
        pred_mark = self.predict_latest(md_df)
        is_buy, is_sell = pred_mark == 1, pred_mark == 0
        # trade_date = md_df['trade_date'].iloc[-1]
        # logger.info('%s is_buy=%s, is_sell=%s', trade_date, str(is_buy), str(is_sell))
        close = md_df['close'].iloc[-1]
        instrument_id = context[ContextKey.instrument_id_list][0]
        if is_buy:  # is_buy
            position_date_pos_info_dic = self.get_position(instrument_id)
            no_target_position = True
            if position_date_pos_info_dic is not None:
                for position_date, pos_info in position_date_pos_info_dic.items():
                    direction = pos_info.direction
                    if direction == Direction.Short:
                        self.close_short(instrument_id, close, pos_info.position)
                    elif direction == Direction.Long:
                        no_target_position = False
            if no_target_position:
                self.open_long(instrument_id, close, self.unit)
            else:
                logger.debug("%s %s     %.2f holding", self.trade_agent.curr_timestamp, instrument_id, close)

        if is_sell:  # is_sell
            position_date_pos_info_dic = self.get_position(instrument_id)
            no_holding_target_position = True
            if position_date_pos_info_dic is not None:
                for position_date, pos_info in position_date_pos_info_dic.items():
                    direction = pos_info.direction
                    if direction == Direction.Long:
                        self.close_long(instrument_id, close, pos_info.position)
                    elif direction == Direction.Short:
                        no_holding_target_position = False
            if no_holding_target_position:
                self.open_short(instrument_id, close, self.unit)
            else:
                logger.debug("%s %s     %.2f holding", self.trade_agent.curr_timestamp, instrument_id, close)

    def separate_train_validation(self, factors, labels):
        """
        将 结果按照比例拆分成训练集、验证集
        :param factors:
        :param labels:
        :return:
        """
        data_len = factors.shape[0]
        train_len = int(self.train_validation_rate * data_len)
        # 至少留给 validation 样本集一个 batch_size 的数量
        if train_len > (data_len - self.batch_size - self.n_step):
            train_len = data_len - self.batch_size - self.n_step

        factors_train, factors_validation = factors[:train_len, :], factors[train_len:, :]
        labels_train, labels_validation = labels[:train_len, :], labels[train_len:, :]
        return factors_train, factors_validation, labels_train, labels_validation

    def model_file_exists(self):
        folder_path, file_name = os.path.split(self.model_file_path)
        if not os.path.exists(folder_path):
            return False
        for f_name in os.listdir(folder_path):
            if f_name.find(file_name) == 0:
                return True

        return False


def _test_use(is_plot):
    from ibats_common import module_root_path
    import os
    # 参数设置
    run_mode = RunMode.Backtest
    strategy_params = {'unit': 1}
    md_agent_params_list = [{
        'md_period': PeriodType.Min1,
        'instrument_id_list': ['RB'],
        'datetime_key': 'trade_date',
        'init_md_date_from': '1995-1-1',  # 行情初始化加载历史数据，供策略分析预加载使用
        'init_md_date_to': '2013-1-1',
        # 'C:\GitHub\IBATS_Common\ibats_common\example\ru_price2.csv'
        'file_path': os.path.abspath(os.path.join(module_root_path, 'example', 'data', 'RB.csv')),
        'symbol_key': 'instrument_type',
    }]
    if run_mode == RunMode.Realtime:
        trade_agent_params = {
        }
        strategy_handler_param = {
        }
    else:
        trade_agent_params = {
            'trade_mode': BacktestTradeMode.Order_2_Deal,
            'init_cash': 10000,
            "calc_mode": CalcMode.Margin,
        }
        strategy_handler_param = {
            'date_from': '2013-1-1',  # 策略回测历史数据，回测指定时间段的历史行情
            'date_to': '2018-10-18',
        }
    # 初始化策略处理器
    stghandler = strategy_handler_factory(
        stg_class=AIStg,
        strategy_params=strategy_params,
        md_agent_params_list=md_agent_params_list,
        exchange_name=ExchangeName.LocalFile,
        run_mode=RunMode.Backtest,
        trade_agent_params=trade_agent_params,
        strategy_handler_param=strategy_handler_param,
    )
    stghandler.start()
    time.sleep(10)
    stghandler.keep_running = False
    stghandler.join()
    stg_run_id = stghandler.stg_run_id
    logging.info("执行结束 stg_run_id = %d", stg_run_id)

    if is_plot:
        from ibats_common.analysis.plot_db import show_order, show_cash_and_margin, show_rr_with_md
        show_order(stg_run_id)
        show_cash_and_margin(stg_run_id)
        show_rr_with_md(stg_run_id)

    return stg_run_id


if __name__ == '__main__':
    # logging.basicConfig(level=logging.DEBUG, format=config.LOG_FORMAT)
    is_plot = True
    _test_use(is_plot)



# -*- coding: utf-8 -*-
"""
Created on 2017/6/9
@author: MG
"""
import logging
from ibats_common.config import ConfigBase as ConBase, update_db_config
from ibats_common.common import ExchangeName

logger = logging.getLogger()


class ConfigBase(ConBase):
    # 交易所名称
    MARKET_NAME = ExchangeName.LocalFile.name

    def __init__(self):
        super().__init__()
        # 设定默认日志格式
        logging.basicConfig(level=logging.DEBUG, format=self.LOG_FORMAT)
        # 设置rest调用日志输出级别
        logging.getLogger('EventAgent').setLevel(logging.INFO)
        logging.getLogger('StgBase').setLevel(logging.INFO)


# 开发配置（SIMNOW MD + Trade）
config = ConfigBase()
update_db_config(config.DB_URL_DIC)


def update_config(config_new: ConfigBase, update_db=True):
    global config
    config = config_new
    logger.info('更新默认配置信息 %s < %s', ConfigBase, config_new.__class__)
    if update_db:
        update_db_config(config.DB_URL_DIC)



#! /usr/bin/env python
# -*- coding:utf-8 -*-
"""
@author  : MG
@Time    : 19-4-24 上午7:55
@File    : mess.py
@contact : mmmaaaggg@163.com
@desc    : 
"""
import pandas as pd
from sqlalchemy.sql import func
from ibats_utils.db import with_db_session, get_db_session, execute_scalar
from ibats_common.backend.orm import StgRunStatusDetail, OrderDetail, TradeDetail, StgRunInfo
from ibats_common.backend import engines


def csv_formatter(file_path_from, file_path_to):
    """
    将 data_integration_celery/tasks/wind/future_reorg/reorg_md_2_db.py 导出的文件重新格式化
    原始文件包括的字段有：
    trade_date,Contract,ContractNext,Close,CloseNext,TermStructure,Volume,VolumeNext,OI,OINext,
    Open,OpenNext,High,HighNext,Low,LowNext,WarehouseWarrant,WarehouseWarrantNext,
    adj_factor_main,adj_factor_secondary,instrument_type
    :param file_path_from:
    :param file_path_to:
    :return:
    """
    df = pd.read_csv(file_path_from)
    col_name_list = ["instrument_type", "trade_date", "Open", "High", "Low", "Close", "Volume",
                     "OI", "WarehouseWarrant", "TermStructure"]
    ret_df = df[col_name_list].rename(columns={_: _.lower() for _ in col_name_list})
    ret_df.to_csv(file_path_to, index=False)


def get_stg_run_info(stg_run_id=None) -> StgRunInfo:
    """
    查询数据库获取最新的 stg_run_id
    :param stg_run_id:
    :return:
    """
    engine_ibats = engines.engine_ibats
    # 获取 收益曲线
    with with_db_session(engine_ibats) as session:
        if stg_run_id is None:
            stg_run_id = session.query(func.max(StgRunInfo.stg_run_id)).scalar()

        info = session.query(StgRunInfo).filter(StgRunInfo.stg_run_id == stg_run_id).first()

    return info


def get_stg_run_id_latest():
    """获取最新的 stg_run_id"""
    engine_ibats = engines.engine_ibats
    with with_db_session(engine_ibats) as session:
        stg_run_id = session.query(func.max(StgRunInfo.stg_run_id)).scalar()

    return stg_run_id


if __name__ == "__main__":
    import os
    folder_path_from = "/home/mg/Downloads/commodity_daily"
    file_name = "RB ConInfoFull_Adj.csv"
    file_path_from = os.path.join(folder_path_from, file_name)
    folder_path_to = "/home/mg/github/IBATS_Common/ibats_common/example/data"
    file_path_to = os.path.join(folder_path_to, f"{file_name.split(' ')[0]}.csv")
    csv_formatter(file_path_from, file_path_to)
